# SPDX-License-Identifier: MIT
from dataclasses import dataclass, field
from itertools import chain
from typing import Any
from xml.etree import ElementTree

from .admindata import AdminData
from .dataobjectproperty import DataObjectProperty
from .dopbase import DopBase
from .dtcdop import DtcDop
from .dynamicendmarkerfield import DynamicEndmarkerField
from .dynamiclengthfield import DynamicLengthField
from .endofpdufield import EndOfPduField
from .environmentdata import EnvironmentData
from .environmentdatadescription import EnvironmentDataDescription
from .multiplexer import Multiplexer
from .nameditemlist import NamedItemList
from .odxdoccontext import OdxDocContext
from .odxlink import OdxLinkDatabase, OdxLinkId
from .snrefcontext import SnRefContext
from .specialdatagroup import SpecialDataGroup
from .staticfield import StaticField
from .structure import Structure
from .table import Table
from .unitspec import UnitSpec


@dataclass(kw_only=True)
class DiagDataDictionarySpec:
    admin_data: AdminData | None = None
    dtc_dops: NamedItemList[DtcDop] = field(default_factory=NamedItemList)
    env_data_descs: NamedItemList[EnvironmentDataDescription] = field(default_factory=NamedItemList)
    data_object_props: NamedItemList[DataObjectProperty] = field(default_factory=NamedItemList)
    structures: NamedItemList[Structure] = field(default_factory=NamedItemList)
    static_fields: NamedItemList[StaticField] = field(default_factory=NamedItemList)
    dynamic_length_fields: NamedItemList[DynamicLengthField] = field(default_factory=NamedItemList)
    dynamic_endmarker_fields: NamedItemList[DynamicEndmarkerField] = field(
        default_factory=NamedItemList)
    end_of_pdu_fields: NamedItemList[EndOfPduField] = field(default_factory=NamedItemList)
    muxs: NamedItemList[Multiplexer] = field(default_factory=NamedItemList)
    env_datas: NamedItemList[EnvironmentData] = field(default_factory=NamedItemList)
    unit_spec: UnitSpec | None = None
    tables: NamedItemList[Table] = field(default_factory=NamedItemList)
    sdgs: list[SpecialDataGroup] = field(default_factory=list)

    @staticmethod
    def from_et(et_element: ElementTree.Element,
                context: OdxDocContext) -> "DiagDataDictionarySpec":
        admin_data = None
        if (admin_data_elem := et_element.find("ADMIN-DATA")) is not None:
            admin_data = AdminData.from_et(admin_data_elem, context)

        dtc_dops = NamedItemList([
            DtcDop.from_et(dtc_dop_elem, context)
            for dtc_dop_elem in et_element.iterfind("DTC-DOPS/DTC-DOP")
        ])

        env_data_descs = NamedItemList([
            EnvironmentDataDescription.from_et(env_data_desc_element, context)
            for env_data_desc_element in et_element.iterfind("ENV-DATA-DESCS/ENV-DATA-DESC")
        ])

        data_object_props = NamedItemList([
            DataObjectProperty.from_et(dop_element, context)
            for dop_element in et_element.iterfind("DATA-OBJECT-PROPS/DATA-OBJECT-PROP")
        ])

        structures = NamedItemList([
            Structure.from_et(structure_element, context)
            for structure_element in et_element.iterfind("STRUCTURES/STRUCTURE")
        ])

        static_fields = NamedItemList([
            StaticField.from_et(dl_element, context)
            for dl_element in et_element.iterfind("STATIC-FIELDS/STATIC-FIELD")
        ])

        dynamic_length_fields = NamedItemList([
            DynamicLengthField.from_et(dl_element, context)
            for dl_element in et_element.iterfind("DYNAMIC-LENGTH-FIELDS/DYNAMIC-LENGTH-FIELD")
        ])

        dynamic_endmarker_fields = NamedItemList([
            DynamicEndmarkerField.from_et(dl_element, context) for dl_element in
            et_element.iterfind("DYNAMIC-ENDMARKER-FIELDS/DYNAMIC-ENDMARKER-FIELD")
        ])

        end_of_pdu_fields = NamedItemList([
            EndOfPduField.from_et(eofp_element, context)
            for eofp_element in et_element.iterfind("END-OF-PDU-FIELDS/END-OF-PDU-FIELD")
        ])

        muxs = NamedItemList([
            Multiplexer.from_et(mux_element, context)
            for mux_element in et_element.iterfind("MUXS/MUX")
        ])

        env_data_elements = chain(
            et_element.iterfind("ENV-DATAS/ENV-DATA"),
            # ODX 2.0.0 says ENV-DATA-DESC could contain a list of ENV-DATAS
            et_element.iterfind("ENV-DATA-DESCS/ENV-DATA-DESC/ENV-DATAS/ENV-DATA"),
        )
        env_datas = NamedItemList([
            EnvironmentData.from_et(env_data_element, context)
            for env_data_element in env_data_elements
        ])

        if (spec_elem := et_element.find("UNIT-SPEC")) is not None:
            unit_spec = UnitSpec.from_et(spec_elem, context)
        else:
            unit_spec = None

        tables = NamedItemList([
            Table.from_et(table_element, context)
            for table_element in et_element.iterfind("TABLES/TABLE")
        ])

        sdgs = [SpecialDataGroup.from_et(sdge, context) for sdge in et_element.iterfind("SDGS/SDG")]

        return DiagDataDictionarySpec(
            admin_data=admin_data,
            dtc_dops=dtc_dops,
            env_data_descs=env_data_descs,
            data_object_props=data_object_props,
            structures=structures,
            static_fields=static_fields,
            dynamic_length_fields=dynamic_length_fields,
            dynamic_endmarker_fields=dynamic_endmarker_fields,
            end_of_pdu_fields=end_of_pdu_fields,
            muxs=muxs,
            env_datas=env_datas,
            unit_spec=unit_spec,
            tables=tables,
            sdgs=sdgs,
        )

    def __post_init__(self) -> None:
        self._all_data_object_properties: NamedItemList[DopBase] = NamedItemList(
            chain(
                self.dtc_dops,
                self.env_data_descs,
                self.data_object_props,
                self.structures,
                self.static_fields,
                self.dynamic_length_fields,
                self.dynamic_endmarker_fields,
                self.end_of_pdu_fields,
                self.muxs,
                self.env_datas,
            ))

    def _build_odxlinks(self) -> dict[OdxLinkId, Any]:
        # note that DataDictionarySpec objects do not exhibit an ODXLINK id.
        odxlinks = {}

        if self.admin_data is not None:
            odxlinks.update(self.admin_data._build_odxlinks())
        for dtc_dop in self.dtc_dops:
            odxlinks.update(dtc_dop._build_odxlinks())
        for env_data_desc in self.env_data_descs:
            odxlinks.update(env_data_desc._build_odxlinks())
        for data_object_prop in self.data_object_props:
            odxlinks.update(data_object_prop._build_odxlinks())
        for structure in self.structures:
            odxlinks.update(structure._build_odxlinks())
        for static_field in self.static_fields:
            odxlinks.update(static_field._build_odxlinks())
        for dynamic_length_field in self.dynamic_length_fields:
            odxlinks.update(dynamic_length_field._build_odxlinks())
        for dynamic_endmarker_field in self.dynamic_endmarker_fields:
            odxlinks.update(dynamic_endmarker_field._build_odxlinks())
        for end_of_pdu_field in self.end_of_pdu_fields:
            odxlinks.update(end_of_pdu_field._build_odxlinks())
        for mux in self.muxs:
            odxlinks.update(mux._build_odxlinks())
        for env_data in self.env_datas:
            odxlinks.update(env_data._build_odxlinks())
        if self.unit_spec is not None:
            odxlinks.update(self.unit_spec._build_odxlinks())
        for table in self.tables:
            odxlinks.update(table._build_odxlinks())
        for sdg in self.sdgs:
            odxlinks.update(sdg._build_odxlinks())

        return odxlinks

    def _resolve_odxlinks(self, odxlinks: OdxLinkDatabase) -> None:
        if self.admin_data is not None:
            self.admin_data._resolve_odxlinks(odxlinks)
        for dtc_dop in self.dtc_dops:
            dtc_dop._resolve_odxlinks(odxlinks)
        for env_data_desc in self.env_data_descs:
            env_data_desc._resolve_odxlinks(odxlinks)
        for data_object_prop in self.data_object_props:
            data_object_prop._resolve_odxlinks(odxlinks)
        for structure in self.structures:
            structure._resolve_odxlinks(odxlinks)
        for static_field in self.static_fields:
            static_field._resolve_odxlinks(odxlinks)
        for dynamic_length_field in self.dynamic_length_fields:
            dynamic_length_field._resolve_odxlinks(odxlinks)
        for dynamic_endmarker_field in self.dynamic_endmarker_fields:
            dynamic_endmarker_field._resolve_odxlinks(odxlinks)
        for end_of_pdu_field in self.end_of_pdu_fields:
            end_of_pdu_field._resolve_odxlinks(odxlinks)
        for mux in self.muxs:
            mux._resolve_odxlinks(odxlinks)
        for env_data in self.env_datas:
            env_data._resolve_odxlinks(odxlinks)
        if self.unit_spec is not None:
            self.unit_spec._resolve_odxlinks(odxlinks)
        for table in self.tables:
            table._resolve_odxlinks(odxlinks)
        for sdg in self.sdgs:
            sdg._resolve_odxlinks(odxlinks)

    def _resolve_snrefs(self, context: SnRefContext) -> None:
        if self.admin_data is not None:
            self.admin_data._resolve_snrefs(context)
        for dtc_dop in self.dtc_dops:
            dtc_dop._resolve_snrefs(context)
        for env_data_desc in self.env_data_descs:
            env_data_desc._resolve_snrefs(context)
        for data_object_prop in self.data_object_props:
            data_object_prop._resolve_snrefs(context)
        for structure in self.structures:
            structure._resolve_snrefs(context)
        for static_field in self.static_fields:
            static_field._resolve_snrefs(context)
        for dynamic_length_field in self.dynamic_length_fields:
            dynamic_length_field._resolve_snrefs(context)
        for dynamic_endmarker_field in self.dynamic_endmarker_fields:
            dynamic_endmarker_field._resolve_snrefs(context)
        for end_of_pdu_field in self.end_of_pdu_fields:
            end_of_pdu_field._resolve_snrefs(context)
        for mux in self.muxs:
            mux._resolve_snrefs(context)
        for env_data in self.env_datas:
            env_data._resolve_snrefs(context)
        if self.unit_spec is not None:
            self.unit_spec._resolve_snrefs(context)
        for table in self.tables:
            table._resolve_snrefs(context)
        for sdg in self.sdgs:
            sdg._resolve_snrefs(context)

    @property
    def all_data_object_properties(self) -> NamedItemList[DopBase]:
        return self._all_data_object_properties



# SPDX-License-Identifier: MIT
from dataclasses import dataclass
from typing import Any
from xml.etree import ElementTree

from deprecation import deprecated

from .dopbase import DopBase
from .element import IdentifiableElement
from .exceptions import odxrequire
from .odxdoccontext import OdxDocContext
from .odxlink import OdxLinkDatabase, OdxLinkId, OdxLinkRef
from .snrefcontext import SnRefContext
from .utils import dataclass_fields_asdict


@dataclass(kw_only=True)
class OutputParam(IdentifiableElement):
    dop_base_ref: OdxLinkRef
    semantic: str | None = None

    @property
    def dop(self) -> DopBase:
        return self._dop

    @deprecated(details="use .dop")  # type: ignore[misc]
    def dop_base(self) -> DopBase:
        return self._dop

    @staticmethod
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "OutputParam":

        kwargs = dataclass_fields_asdict(IdentifiableElement.from_et(et_element, context))

        dop_base_ref = odxrequire(OdxLinkRef.from_et(et_element.find("DOP-BASE-REF"), context))
        semantic = et_element.get("SEMANTIC")

        return OutputParam(dop_base_ref=dop_base_ref, semantic=semantic, **kwargs)

    def _build_odxlinks(self) -> dict[OdxLinkId, Any]:
        return {}

    def _resolve_odxlinks(self, odxlinks: OdxLinkDatabase) -> None:
        self._dop = odxlinks.resolve(self.dop_base_ref, DopBase)

    def _resolve_snrefs(self, context: SnRefContext) -> None:
        pass



# SPDX-License-Identifier: MIT
from dataclasses import dataclass, field
from typing import Any
from xml.etree import ElementTree

from ..basevariantpattern import BaseVariantPattern
from ..diagvariable import DiagVariable
from ..dyndefinedspec import DynDefinedSpec
from ..exceptions import odxraise
from ..nameditemlist import NamedItemList
from ..odxdoccontext import OdxDocContext
from ..odxlink import OdxLinkDatabase, OdxLinkId, OdxLinkRef
from ..parentref import ParentRef
from ..snrefcontext import SnRefContext
from ..utils import dataclass_fields_asdict
from ..variablegroup import VariableGroup
from .hierarchyelementraw import HierarchyElementRaw


@dataclass(kw_only=True)
class BaseVariantRaw(HierarchyElementRaw):
    """This is a diagnostic layer for common functionality of an ECU
    """

    diag_variables_raw: list[DiagVariable | OdxLinkRef] = field(default_factory=list)
    variable_groups: NamedItemList[VariableGroup] = field(default_factory=NamedItemList)
    dyn_defined_spec: DynDefinedSpec | None = None
    base_variant_pattern: BaseVariantPattern | None = None
    parent_refs: list[ParentRef] = field(default_factory=list)

    @property
    def diag_variables(self) -> NamedItemList[DiagVariable]:
        return self._diag_variables

    @staticmethod
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "BaseVariantRaw":
        # objects contained by diagnostic layers exhibit an additional
        # document fragment for the diag layer, so we use the document
        # fragments of the odx id of the diag layer for IDs of
        # contained objects.
        her = HierarchyElementRaw.from_et(et_element, context)
        kwargs = dataclass_fields_asdict(her)

        diag_variables_raw: list[DiagVariable | OdxLinkRef] = []
        if (dv_elems := et_element.find("DIAG-VARIABLES")) is not None:
            for dv_proxy_elem in dv_elems:
                dv_proxy: OdxLinkRef | DiagVariable
                if dv_proxy_elem.tag == "DIAG-VARIABLE-REF":
                    dv_proxy = OdxLinkRef.from_et(dv_proxy_elem, context)
                elif dv_proxy_elem.tag == "DIAG-VARIABLE":
                    dv_proxy = DiagVariable.from_et(dv_proxy_elem, context)
                else:
                    odxraise()

                diag_variables_raw.append(dv_proxy)

        variable_groups = NamedItemList([
            VariableGroup.from_et(vg_elem, context)
            for vg_elem in et_element.iterfind("VARIABLE-GROUPS/VARIABLE-GROUP")
        ])

        dyn_defined_spec = None
        if (dds_elem := et_element.find("DYN-DEFINED-SPEC")) is not None:
            dyn_defined_spec = DynDefinedSpec.from_et(dds_elem, context)

        base_variant_pattern = None
        if (bvp_elem := et_element.find("BASE-VARIANT-PATTERN")) is not None:
            base_variant_pattern = BaseVariantPattern.from_et(bvp_elem, context)

        parent_refs = [
            ParentRef.from_et(pr_elem, context)
            for pr_elem in et_element.iterfind("PARENT-REFS/PARENT-REF")
        ]

        return BaseVariantRaw(
            diag_variables_raw=diag_variables_raw,
            variable_groups=variable_groups,
            dyn_defined_spec=dyn_defined_spec,
            base_variant_pattern=base_variant_pattern,
            parent_refs=parent_refs,
            **kwargs)

    def _build_odxlinks(self) -> dict[OdxLinkId, Any]:
        result = super()._build_odxlinks()

        for dv_proxy in self.diag_variables_raw:
            if not isinstance(dv_proxy, OdxLinkRef):
                result.update(dv_proxy._build_odxlinks())

        if self.dyn_defined_spec is not None:
            result.update(self.dyn_defined_spec._build_odxlinks())

        for parent_ref in self.parent_refs:
            result.update(parent_ref._build_odxlinks())

        return result

    def _resolve_odxlinks(self, odxlinks: OdxLinkDatabase) -> None:
        super()._resolve_odxlinks(odxlinks)

        self._diag_variables: NamedItemList[DiagVariable] = NamedItemList()
        for dv_proxy in self.diag_variables_raw:
            if isinstance(dv_proxy, OdxLinkRef):
                dv = odxlinks.resolve(dv_proxy, DiagVariable)
            else:
                dv_proxy._resolve_odxlinks(odxlinks)
                dv = dv_proxy

            self._diag_variables.append(dv)

        if self.dyn_defined_spec is not None:
            self.dyn_defined_spec._resolve_odxlinks(odxlinks)

        for parent_ref in self.parent_refs:
            parent_ref._resolve_odxlinks(odxlinks)

    def _resolve_snrefs(self, context: SnRefContext) -> None:
        super()._resolve_snrefs(context)

        for dv_proxy in self.diag_variables_raw:
            if not isinstance(dv_proxy, OdxLinkRef):
                dv_proxy._resolve_snrefs(context)

        if self.dyn_defined_spec is not None:
            self.dyn_defined_spec._resolve_snrefs(context)

        for parent_ref in self.parent_refs:
            parent_ref._resolve_snrefs(context)



# SPDX-License-Identifier: MIT
import json
from typing import Any

import pytest

from odxtools.basevariantpattern import BaseVariantPattern
from odxtools.database import Database
from odxtools.diaglayers.basevariant import BaseVariant
from odxtools.diaglayers.basevariantraw import BaseVariantRaw
from odxtools.diaglayers.diaglayertype import DiagLayerType
from odxtools.diaglayers.ecuvariant import EcuVariant
from odxtools.diaglayers.ecuvariantraw import EcuVariantRaw
from odxtools.diagservice import DiagService
from odxtools.ecuvariantpattern import EcuVariantPattern
from odxtools.exceptions import OdxError, odxrequire
from odxtools.matchingbasevariantparameter import MatchingBaseVariantParameter
from odxtools.matchingparameter import MatchingParameter
from odxtools.odxlink import DocType, OdxDocFragment, OdxLinkDatabase, OdxLinkId, OdxLinkRef
from odxtools.request import Request
from odxtools.response import Response, ResponseType
from odxtools.variantmatcher import VariantMatcher

doc_frags = [OdxDocFragment(doc_name="pytest", doc_type=DocType.CONTAINER)]

odxlinks = OdxLinkDatabase()


@pytest.fixture
def dummy_response(monkeypatch: pytest.MonkeyPatch) -> Response:
    resp = Response(
        odx_id=OdxLinkId(local_id="dummy_resp", doc_fragments=doc_frags),
        short_name="dummy_resp",
        response_type=ResponseType.POSITIVE,
    )
    odxlinks.update({resp.odx_id: resp})

    def decode(message: bytes) -> dict[str, Any]:
        msg_str = message.decode(encoding="utf-8")
        msg_dict = json.loads(msg_str)
        assert isinstance(msg_dict, dict)
        return msg_dict

    monkeypatch.setattr(resp, "decode", decode)
    return resp


@pytest.fixture()
def bv_ident_service(monkeypatch: pytest.MonkeyPatch, dummy_response: Response) -> DiagService:
    dummy_req = Request(
        odx_id=OdxLinkId(local_id="dummy_req", doc_fragments=doc_frags),
        short_name="dummy_req",
    )
    odxlinks.update({dummy_req.odx_id: dummy_req})

    diagService = DiagService(
        odx_id=OdxLinkId(local_id="identService", doc_fragments=doc_frags),
        short_name="identService",
        request_ref=OdxLinkRef.from_id(dummy_req.odx_id),
        pos_response_refs=[OdxLinkRef.from_id(dummy_response.odx_id)],
    )

    def encode_request() -> bytes:
        return b"\xff\xee\xdd"

    monkeypatch.setattr(diagService, "encode_request", encode_request)
    return diagService


@pytest.fixture()
def ev_ident_service(monkeypatch: pytest.MonkeyPatch, dummy_response: Response) -> DiagService:
    dummy_req = Request(
        odx_id=OdxLinkId(local_id="dummy_req", doc_fragments=doc_frags),
        short_name="dummy_req",
    )
    odxlinks.update({dummy_req.odx_id: dummy_req})

    diagService = DiagService(
        odx_id=OdxLinkId(local_id="identService", doc_fragments=doc_frags),
        short_name="identService",
        request_ref=OdxLinkRef.from_id(dummy_req.odx_id),
        pos_response_refs=[OdxLinkRef.from_id(dummy_response.odx_id)],
    )

    def encode_request() -> bytes:
        return b"\x22\x10\x00"

    monkeypatch.setattr(diagService, "encode_request", encode_request)
    return diagService


@pytest.fixture
def ev_supplier_service(monkeypatch: pytest.MonkeyPatch, dummy_response: Response) -> DiagService:
    dummy_req = Request(
        odx_id=OdxLinkId(local_id="dummy_req", doc_fragments=doc_frags),
        short_name="dummy_req",
    )
    odxlinks.update({dummy_req.odx_id: dummy_req})

    diagService = DiagService(
        odx_id=OdxLinkId(local_id="supplierService", doc_fragments=doc_frags),
        short_name="supplierService",
        request_ref=OdxLinkRef.from_id(dummy_req.odx_id),
        pos_response_refs=[OdxLinkRef.from_id(dummy_response.odx_id)],
    )

    def encode_request() -> bytes:
        return b"\x22\x20\x00"

    monkeypatch.setattr(diagService, "encode_request", encode_request)
    return diagService


@pytest.fixture
def bv_supplier_service(monkeypatch: pytest.MonkeyPatch, dummy_response: Response) -> DiagService:
    dummy_req = Request(
        odx_id=OdxLinkId(local_id="dummy_req", doc_fragments=doc_frags),
        short_name="dummy_req",
    )
    odxlinks.update({dummy_req.odx_id: dummy_req})

    diagService = DiagService(
        odx_id=OdxLinkId(local_id="supplierService", doc_fragments=doc_frags),
        short_name="supplierService",
        request_ref=OdxLinkRef.from_id(dummy_req.odx_id),
        pos_response_refs=[OdxLinkRef.from_id(dummy_response.odx_id)],
    )

    def encode_request() -> bytes:
        return b"\xcc\xbb\xaa"

    monkeypatch.setattr(diagService, "encode_request", encode_request)
    return diagService


@pytest.fixture
def bv2_supplier_service(monkeypatch: pytest.MonkeyPatch, dummy_response: Response) -> DiagService:
    bv2_req = Request(
        odx_id=OdxLinkId(local_id="bv2_req", doc_fragments=doc_frags),
        short_name="bv2_req",
    )
    odxlinks.update({bv2_req.odx_id: bv2_req})

    diagService = DiagService(
        odx_id=OdxLinkId(local_id="bv2.supplierService", doc_fragments=doc_frags),
        short_name="supplierService",
        request_ref=OdxLinkRef.from_id(bv2_req.odx_id),
        pos_response_refs=[OdxLinkRef.from_id(dummy_response.odx_id)],
    )

    def encode_request() -> bytes:
        return b"\xcc\xbb\xab"

    monkeypatch.setattr(diagService, "encode_request", encode_request)
    return diagService


@pytest.fixture
def base_variant_pattern1() -> BaseVariantPattern:
    return BaseVariantPattern(matching_base_variant_parameters=[
        MatchingBaseVariantParameter(
            diag_comm_snref="identService",
            expected_value="1234",
            out_param_if_snref="id",
        ),
        MatchingBaseVariantParameter(
            diag_comm_snref="supplierService",
            use_physical_addressing_raw=False,
            expected_value="OEM",
            out_param_if_snpathref="info.type",
        ),
    ])


@pytest.fixture
def base_variant_pattern2() -> BaseVariantPattern:
    return BaseVariantPattern(matching_base_variant_parameters=[
        MatchingBaseVariantParameter(
            diag_comm_snref="identService",
            expected_value="1234",
            out_param_if_snref="id",
        ),
        MatchingBaseVariantParameter(
            diag_comm_snref="supplierService",
            use_physical_addressing_raw=True,
            expected_value="tier1",
            out_param_if_snpathref="info.type",
        ),
    ])


@pytest.fixture
def base_variant_pattern3() -> BaseVariantPattern:
    return BaseVariantPattern(matching_base_variant_parameters=[
        MatchingBaseVariantParameter(
            diag_comm_snref="supplierService",
            use_physical_addressing_raw=False,
            expected_value="tier5",
            out_param_if_snpathref="info.type",
        )
    ])


@pytest.fixture
def ecu_variant_pattern1() -> EcuVariantPattern:
    return EcuVariantPattern(matching_parameters=[
        MatchingParameter(
            diag_comm_snref="identService",
            expected_value="1000",
            out_param_if_snref="id",
        ),
        MatchingParameter(
            diag_comm_snref="supplierService",
            expected_value="supplier_A",
            out_param_if_snpathref="name.english",
        ),
    ])


@pytest.fixture
def ecu_variant_pattern2() -> EcuVariantPattern:
    return EcuVariantPattern(matching_parameters=[
        MatchingParameter(
            diag_comm_snref="identService",
            expected_value="2000",
            out_param_if_snref="id",
        ),
        MatchingParameter(
            diag_comm_snref="supplierService",
            expected_value="supplier_B",
            out_param_if_snpathref="name.english",
        ),
    ])


@pytest.fixture
def ecu_variant_pattern3() -> EcuVariantPattern:
    return EcuVariantPattern(matching_parameters=[
        MatchingParameter(
            diag_comm_snref="supplierService",
            expected_value="supplier_C",
            out_param_if_snpathref="name.english",
        )
    ])


@pytest.fixture
def base_variant_1(
    bv_ident_service: DiagService,
    bv_supplier_service: DiagService,
    base_variant_pattern1: BaseVariantPattern,
) -> BaseVariant:
    raw_layer = BaseVariantRaw(
        variant_type=DiagLayerType.BASE_VARIANT,
        odx_id=OdxLinkId(local_id="base_variant1", doc_fragments=doc_frags),
        short_name="base_variant1",
        diag_comms_raw=[bv_ident_service, bv_supplier_service],
        base_variant_pattern=base_variant_pattern1,
    )
    result = BaseVariant(diag_layer_raw=raw_layer)
    odxlinks.update(result._build_odxlinks())
    db = Database()
    result._resolve_odxlinks(odxlinks)
    result._finalize_init(db, odxlinks)
    return result


@pytest.fixture
def base_variant_2(
    bv_ident_service: DiagService,
    bv2_supplier_service: DiagService,
    base_variant_pattern2: BaseVariantPattern,
) -> BaseVariant:
    raw_layer = BaseVariantRaw(
        variant_type=DiagLayerType.BASE_VARIANT,
        odx_id=OdxLinkId(local_id="base_variant2", doc_fragments=doc_frags),
        short_name="base_variant2",
        diag_comms_raw=[bv_ident_service, bv2_supplier_service],
        base_variant_pattern=base_variant_pattern2,
    )
    result = BaseVariant(diag_layer_raw=raw_layer)
    odxlinks.update(result._build_odxlinks())
    db = Database()
    result._resolve_odxlinks(odxlinks)
    result._finalize_init(db, odxlinks)
    return result


@pytest.fixture
def base_variant_3(
    bv_ident_service: DiagService,
    bv_supplier_service: DiagService,
    base_variant_pattern1: BaseVariantPattern,
    base_variant_pattern3: BaseVariantPattern,
) -> BaseVariant:
    raw_layer = BaseVariantRaw(
        variant_type=DiagLayerType.BASE_VARIANT,
        odx_id=OdxLinkId(local_id="base_variant3", doc_fragments=doc_frags),
        short_name="base_variant3",
        diag_comms_raw=[bv_ident_service, bv_supplier_service],
        base_variant_pattern=base_variant_pattern3,
    )
    result = BaseVariant(diag_layer_raw=raw_layer)
    odxlinks.update(result._build_odxlinks())
    db = Database()
    result._resolve_odxlinks(odxlinks)
    result._finalize_init(db, odxlinks)
    return result


@pytest.fixture
def base_variants(base_variant_1: BaseVariant, base_variant_2: BaseVariant,
                  base_variant_3: BaseVariant) -> list[BaseVariant]:
    return [base_variant_1, base_variant_2, base_variant_3]


@pytest.fixture
def ecu_variant_1(
    ev_ident_service: DiagService,
    ev_supplier_service: DiagService,
    ecu_variant_pattern1: EcuVariantPattern,
) -> EcuVariant:
    raw_layer = EcuVariantRaw(
        variant_type=DiagLayerType.ECU_VARIANT,
        odx_id=OdxLinkId(local_id="ecu_variant1", doc_fragments=doc_frags),
        short_name="ecu_variant1",
        diag_comms_raw=[ev_ident_service, ev_supplier_service],
        ecu_variant_patterns=[ecu_variant_pattern1],
    )
    result = EcuVariant(diag_layer_raw=raw_layer)
    odxlinks.update(result._build_odxlinks())
    db = Database()
    result._resolve_odxlinks(odxlinks)
    result._finalize_init(db, odxlinks)
    return result


@pytest.fixture
def ecu_variant_2(
    ev_ident_service: DiagService,
    ev_supplier_service: DiagService,
    ecu_variant_pattern2: EcuVariantPattern,
) -> EcuVariant:
    raw_layer = EcuVariantRaw(
        variant_type=DiagLayerType.ECU_VARIANT,
        odx_id=OdxLinkId(local_id="ecu_variant2", doc_fragments=doc_frags),
        short_name="ecu_variant2",
        diag_comms_raw=[ev_ident_service, ev_supplier_service],
        ecu_variant_patterns=[ecu_variant_pattern2],
    )
    result = EcuVariant(diag_layer_raw=raw_layer)
    odxlinks.update(result._build_odxlinks())
    db = Database()
    result._resolve_odxlinks(odxlinks)
    result._finalize_init(db, odxlinks)
    return result


@pytest.fixture
def ecu_variant_3(
    ev_ident_service: DiagService,
    ev_supplier_service: DiagService,
    ecu_variant_pattern1: EcuVariantPattern,
    ecu_variant_pattern3: EcuVariantPattern,
) -> EcuVariant:
    raw_layer = EcuVariantRaw(
        variant_type=DiagLayerType.ECU_VARIANT,
        odx_id=OdxLinkId(local_id="ecu_variant3", doc_fragments=doc_frags),
        short_name="ecu_variant3",
        diag_comms_raw=[ev_ident_service, ev_supplier_service],
        ecu_variant_patterns=[ecu_variant_pattern1, ecu_variant_pattern3],
    )
    result = EcuVariant(diag_layer_raw=raw_layer)
    odxlinks.update(result._build_odxlinks())
    db = Database()
    result._resolve_odxlinks(odxlinks)
    result._finalize_init(db, odxlinks)
    return result


@pytest.fixture
def ecu_variants(ecu_variant_1: EcuVariant, ecu_variant_2: EcuVariant,
                 ecu_variant_3: EcuVariant) -> list[EcuVariant]:
    return [ecu_variant_1, ecu_variant_2, ecu_variant_3]


def as_bytes(dikt: dict[str, Any]) -> bytes:
    return bytes(json.dumps(dikt), "utf-8")


@pytest.mark.parametrize("use_cache", [True, False])
# the req_resp_mapping maps request to responses for the ecu-under-test
@pytest.mark.parametrize(
    "req_resp_mapping, expected_variant",
    [
        # test if full match of matching parameters is accepted
        (
            {
                b"\xff\xee\xdd": as_bytes({"id": 1234}),
                b"\xcc\xbb\xaa": as_bytes({"info": {
                    "type": "OEM"
                }}),
            },
            "base_variant1",
        ),
        (
            {
                b"\xff\xee\xdd": as_bytes({"id": 1234}),
                b"\xcc\xbb\xab": as_bytes({"info": {
                    "type": "tier1"
                }}),
            },
            "base_variant2",
        ),
    ],
)
def test_base_variant_matching(
    base_variants: list[BaseVariant],
    use_cache: bool,
    req_resp_mapping: dict[bytes, bytes],
    expected_variant: str,
) -> None:

    matcher = VariantMatcher(
        variant_candidates=base_variants,
        use_cache=use_cache,
    )
    has_physical_addressing = False
    has_functional_addressing = False
    for use_physical_addressing, req in matcher.request_loop():
        has_physical_addressing = has_physical_addressing or use_physical_addressing
        has_functional_addressing = has_functional_addressing or not use_physical_addressing
        resp = req_resp_mapping.get(req)
        if resp is not None:
            matcher.evaluate(resp)
        else:
            # we don't know about the request. report back an negative response
            matcher.evaluate(b'{ "SID": 127 }')

    assert has_physical_addressing and has_functional_addressing
    assert matcher.has_match()
    assert odxrequire(matcher.matching_variant).short_name == expected_variant


@pytest.mark.parametrize("use_cache", [True, False])
# the req_resp_mapping maps request to responses for the ecu-under-test
@pytest.mark.parametrize(
    "req_resp_mapping, expected_variant",
    [
        # test if full match of matching parameters is accepted
        (
            {
                b"\x22\x10\x00": as_bytes({"id": 2000}),
                b"\x22\x20\x00": as_bytes({"name": {
                    "english": "supplier_B"
                }}),
            },
            "ecu_variant2",
        ),
        # test if partial match of matching parameters is rejected
        (
            {
                b"\x22\x10\x00": as_bytes({"id": 2000}),
                b"\x22\x20\x00": as_bytes({"name": {
                    "english": "supplier_C"
                }}),
            },
            "ecu_variant3",
        ),
        # test if first full match is preferred over second match
        (
            {
                b"\x22\x10\x00": as_bytes({"id": 1000}),
                b"\x22\x20\x00": as_bytes({"name": {
                    "english": "supplier_A"
                }}),
            },
            "ecu_variant1",
        ),
    ],
)
def test_ecu_variant_matching(
    ecu_variants: list[EcuVariant],
    use_cache: bool,
    req_resp_mapping: dict[bytes, bytes],
    expected_variant: str,
) -> None:
    matcher = VariantMatcher(
        variant_candidates=ecu_variants,
        use_cache=use_cache,
    )
    for use_physical_addressing, req in matcher.request_loop():
        assert use_physical_addressing
        resp = req_resp_mapping[req]
        matcher.evaluate(resp)
    assert matcher.has_match()
    assert odxrequire(matcher.matching_variant).short_name == expected_variant


@pytest.mark.parametrize("use_cache", [True, False])
def test_no_match(ecu_variants: list[EcuVariant], use_cache: bool) -> None:
    # stores the responses for each request for the ecu-under-test
    req_resp_mapping = {
        b"\x22\x10\x00": as_bytes({"id": 1000}),
        b"\x22\x20\x00": as_bytes({"name": {
            "english": "supplier_D"
        }}),
    }

    matcher = VariantMatcher(
        variant_candidates=ecu_variants,
        use_cache=use_cache,
    )
    for use_physical_addressing, req in matcher.request_loop():
        assert use_physical_addressing
        resp = req_resp_mapping[req]
        matcher.evaluate(resp)
    assert not matcher.has_match()
    assert matcher.matching_variant is None


@pytest.mark.parametrize("use_cache", [True, False])
# test if pending matchers reject the has_match() or active variant query
def test_no_request_loop(ecu_variants: list[EcuVariant], use_cache: bool) -> None:
    matcher = VariantMatcher(
        variant_candidates=ecu_variants,
        use_cache=use_cache,
    )
    with pytest.raises(RuntimeError):
        matcher.has_match()
    assert matcher.matching_variant is None


@pytest.mark.parametrize("use_cache", [True, False])
# test if runs of the request loop without calling `evaluate(...)` are rejected
def test_request_loop_misuse(ecu_variants: list[EcuVariant], use_cache: bool) -> None:
    matcher = VariantMatcher(
        variant_candidates=ecu_variants,
        use_cache=use_cache,
    )
    with pytest.raises(RuntimeError):
        for _, _ in matcher.request_loop():
            pass


@pytest.mark.parametrize("use_cache", [True, False])
# test if request loop is idempotent, i.e., the matching is the same regardless of how often the request loop is run
def test_request_loop_idempotency(ecu_variants: list[EcuVariant], use_cache: bool) -> None:
    req_resp_mapping = {
        b"\x22\x10\x00": as_bytes({"id": 2000}),
        b"\x22\x20\x00": as_bytes({"name": {
            "english": "supplier_B"
        }}),
    }

    matcher = VariantMatcher(
        variant_candidates=ecu_variants,
        use_cache=use_cache,
    )

    for _, req in matcher.request_loop():
        resp = req_resp_mapping[req]
        matcher.evaluate(resp)
    assert matcher.has_match()
    assert odxrequire(matcher.matching_variant).short_name == "ecu_variant2"

    # second run with arbitrary input
    for _, _ in matcher.request_loop():
        matcher.evaluate(b"")

    # idempotency criterion
    assert matcher.has_match()
    assert odxrequire(matcher.matching_variant).short_name == "ecu_variant2"


@pytest.mark.parametrize("use_cache", [True, False])
def test_unresolvable_snpathref(ecu_variants: list[EcuVariant], use_cache: bool) -> None:
    # stores the responses for each request for the ecu-under-test
    req_resp_mapping = {
        b"\x22\x10\x00": as_bytes({"id": 1000}),
        # the snpathref cannot be resolved, because name is not a struct
        b"\x22\x20\x00": as_bytes({"name": "supplier_C"}),
    }

    matcher = VariantMatcher(
        variant_candidates=ecu_variants,
        use_cache=use_cache,
    )

    with pytest.raises(OdxError):
        for _, req in matcher.request_loop():
            resp = req_resp_mapping[req]
            matcher.evaluate(resp)



# SPDX-License-Identifier: MIT
from dataclasses import dataclass
from typing import Any
from xml.etree import ElementTree

from .diagnostictroublecode import DiagnosticTroubleCode
from .dtcdop import DtcDop
from .element import NamedElement
from .exceptions import odxrequire
from .odxdoccontext import OdxDocContext
from .odxlink import OdxLinkDatabase, OdxLinkId, OdxLinkRef, resolve_snref
from .snrefcontext import SnRefContext
from .utils import dataclass_fields_asdict


@dataclass(kw_only=True)
class DtcConnector(NamedElement):
    dtc_dop_ref: OdxLinkRef
    dtc_snref: str

    @property
    def dtc_dop(self) -> DtcDop:
        return self._dtc_dop

    @property
    def dtc(self) -> DiagnosticTroubleCode:
        return self._dtc

    @staticmethod
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "DtcConnector":
        kwargs = dataclass_fields_asdict(NamedElement.from_et(et_element, context))

        dtc_dop_ref = odxrequire(OdxLinkRef.from_et(et_element.find("DTC-DOP-REF"), context))
        dtc_snref_el = odxrequire(et_element.find("DTC-SNREF"))
        dtc_snref = odxrequire(dtc_snref_el.get("SHORT-NAME"))

        return DtcConnector(dtc_dop_ref=dtc_dop_ref, dtc_snref=dtc_snref, **kwargs)

    def _build_odxlinks(self) -> dict[OdxLinkId, Any]:
        return {}

    def _resolve_odxlinks(self, odxlinks: OdxLinkDatabase) -> None:
        self._dtc_dop = odxlinks.resolve(self.dtc_dop_ref, DtcDop)

    def _resolve_snrefs(self, context: SnRefContext) -> None:
        self._dtc = resolve_snref(self.dtc_snref, self._dtc_dop.dtcs, DiagnosticTroubleCode)



# SPDX-License-Identifier: MIT
import abc
import typing
from collections.abc import Collection, Iterable
from copy import deepcopy
from keyword import iskeyword
from typing import Any, SupportsIndex, TypeVar, cast, overload, runtime_checkable

from .exceptions import odxraise


@runtime_checkable
class OdxNamed(typing.Protocol):

    @property
    def short_name(self) -> str:
        ...


T = TypeVar("T")
TNamed = TypeVar("TNamed", bound=OdxNamed)


class ItemAttributeList(list[T]):
    """A list that provides direct access to its items as named attributes.

    This is a hybrid between a list and a user-defined object: One can
    iterate over all items of the list as usual, but items can also be
    accessed via `named_list.itemname`, where the 'itemname' is
    specified via a item -> string mapping function that is passed to
    the constructor.

    If an item name is not unique, `_<num>` will be appended to
    avoid naming collisions. The user is responsible that the strings
    returned by the item-to-name function are valid identifiers in python.
    """

    def __init__(self, input_list: Iterable[T] | None = None) -> None:
        self._item_dict: dict[str, T] = {}

        if input_list is not None:
            for item in input_list:
                self.append(item)

    @abc.abstractmethod
    def _get_item_key(self, item: T) -> str:
        pass

    def append(self, item: T) -> None:
        """
        Append a new item to the list and make it accessible as a
        member attribute.

        \return The name under which item is accessible
        """
        self._add_attribute_item(item)

        super().append(item)

    def _add_attribute_item(self, item: T) -> None:
        item_name = self._get_item_key(item)

        # eliminate conflicts between the name of the new item and
        # existing attributes of the ItemAttributeList object
        i = 1
        tmp = item_name
        while True:
            if not hasattr(self, tmp):
                break

            i += 1
            if item_name.endswith("_"):
                # if the item name already ends with an underscore,
                # there's no need to add a second one...
                tmp = f"{item_name}{i}"
            else:
                tmp = f"{item_name}_{i}"

        item_name = tmp

        self._item_dict[item_name] = item

    def insert(self, index: SupportsIndex, obj: T) -> None:
        self._add_attribute_item(obj)

        list.insert(self, index, obj)

    def remove(self, obj: T) -> None:
        list.remove(self, obj)

        keys = [k for (k, v) in self._item_dict.items() if v == obj]
        for key in keys:
            del self._item_dict[key]

    def pop(self, index: SupportsIndex = -1) -> T:
        result = list.pop(self, index)
        keys = [k for (k, v) in self._item_dict.items() if v == result]
        for key in keys:
            del self._item_dict[key]
        return result

    def extend(self, items: Iterable[T]) -> None:
        for item in items:
            self.append(item)

    def clear(self) -> None:
        super().clear()

        self._item_dict = {}

    def copy(self) -> "ItemAttributeList[T]":
        result = self.__class__()
        for item in self:
            list.append(result, item)
        result._item_dict = self._item_dict.copy()
        return result

    def keys(self) -> Collection[str]:
        return self._item_dict.keys()

    def values(self) -> Collection[T]:
        return self._item_dict.values()

    def items(self) -> Collection[tuple[str, T]]:
        return self._item_dict.items()

    def __dir__(self) -> dict[str, Any]:
        result = dict(self.__dict__)
        result.update(self._item_dict)
        return result

    @overload
    def __getitem__(self, key: SupportsIndex) -> T:
        ...

    @overload
    def __getitem__(self, key: str) -> T:
        ...

    @overload
    def __getitem__(self, key: slice) -> list[T]:
        ...

    def __getitem__(self, key: SupportsIndex | str | slice) -> T | list[T]:
        if isinstance(key, (SupportsIndex, slice)):
            return super().__getitem__(key)
        else:
            return self._item_dict[key]

    def __getattr__(self, key: str) -> T:
        if key not in self._item_dict:
            raise AttributeError(f"ItemAttributeList does not contain an item named '{key}'")

        return self._item_dict[key]

    def get(self, key: int | str, default: T | None = None) -> T | None:
        if isinstance(key, int):
            if 0 <= key and key < len(self):
                return super().__getitem__(key)
            return default
        else:
            return cast(T | None, self._item_dict.get(key, default))

    def __eq__(self, other: object) -> bool:
        """
        Item lists are equal if the underlying lists are equal.
        """
        if not isinstance(other, type(self)) or not isinstance(self, type(other)):
            return False
        else:
            return self._item_dict == other._item_dict

    def __str__(self) -> str:
        return f"[{', '.join( [self._get_item_key(x) for x in self])}]"

    def __repr__(self) -> str:
        return f"{type(self).__name__}([{', '.join([repr(x) for x in self])}])"

    def __copy__(self) -> Any:
        return self.__class__(list(self))

    def __deepcopy__(self, memo: dict[int, Any]) -> Any:
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        result._item_dict = {}
        for x in self:
            result.append(deepcopy(x, memo))

        return result

    def __reduce__(self) -> tuple[Any, ...]:
        """Support for Python's pickle protocol.
        This method ensures that the object can be reconstructed with its current state,
        using its class and the list of items it contains.
        It returns a tuple containing the reconstruction function (the class)
        and its arguments necessary to recreate the object.

        """
        return self.__class__, (list(self),)


class NamedItemList(ItemAttributeList[T]):

    def _get_item_key(self, item: T) -> str:
        """Transform an object's `short_name` attribute into a valid
        python identifier

        Although short names are almost identical to valid python
        identifiers, their first character is allowed to be a number or
        they may be python keywords. This method prepends an underscore to
        such short names.

        """
        if not isinstance(item, OdxNamed):
            odxraise()
        sn = item.short_name
        if not isinstance(sn, str):
            odxraise()

        # make sure that the name of the item in question is not a python
        # keyword (this would lead to syntax errors) and that does not
        # start with a digit
        if sn[0].isdigit() or iskeyword(sn):
            return f"_{sn}"

        return sn



# SPDX-License-Identifier: MIT
from enum import Enum

from .exceptions import odxraise
from .odxtypes import DataType


class Encoding(Enum):
    BCD_P = "BCD-P"
    BCD_UP = "BCD-UP"

    ONEC = "1C"
    TWOC = "2C"
    SM = "SM"

    UTF8 = "UTF-8"
    UCS2 = "UCS-2"
    ISO_8859_1 = "ISO-8859-1"
    ISO_8859_2 = "ISO-8859-2"
    WINDOWS_1252 = "WINDOWS-1252"

    NONE = "NONE"


def get_string_encoding(base_data_type: DataType, base_type_encoding: Encoding | None,
                        is_highlow_byte_order: bool) -> str | None:
    """If the encoding is for a string, return the value for
    `str.encode()`/`str.decode()` to convert the string object
    to/from a byte array
    """

    # note that the spec disallows certain combinations of
    # base_data_type and encoding (e.g., A_ASCIISTRING encoded
    # using UTF-8). Since in python3 strings are always
    # capable of the full unicode character set, odxtools
    # ignores these restrictions...
    if base_type_encoding == Encoding.UTF8 or (base_data_type == DataType.A_UTF8STRING and
                                               base_type_encoding is None):
        return "utf-8"
    elif base_type_encoding == Encoding.UCS2 or (base_data_type == DataType.A_UNICODE2STRING and
                                                 base_type_encoding is None):
        return "utf-16-be" if is_highlow_byte_order else "utf-16-le"
    elif base_type_encoding == Encoding.ISO_8859_1 or (base_data_type == DataType.A_ASCIISTRING and
                                                       base_type_encoding is None):
        return "iso-8859-1"
    elif base_type_encoding == Encoding.ISO_8859_2:
        return "iso-8859-2"
    elif base_type_encoding == Encoding.WINDOWS_1252:
        return "cp1252"
    else:
        odxraise(f"Specified illegal encoding {base_type_encoding} for {base_data_type.value} "
                 f"string object")
        return "iso-8859-1"

    return None



# SPDX-License-Identifier: MIT
from dataclasses import dataclass
from typing import Literal
from xml.etree import ElementTree

from typing_extensions import override

from .decodestate import DecodeState
from .diagcodedtype import DctType, DiagCodedType
from .encodestate import EncodeState
from .exceptions import odxassert, odxraise, odxrequire
from .odxdoccontext import OdxDocContext
from .odxtypes import AtomicOdxType, BytesTypes, DataType, odxstr_to_bool
from .utils import dataclass_fields_asdict, read_hex_binary


@dataclass(kw_only=True)
class StandardLengthType(DiagCodedType):

    bit_length: int
    bit_mask: int | None = None
    is_condensed_raw: bool | None = None

    @property
    def dct_type(self) -> DctType:
        return "STANDARD-LENGTH-TYPE"

    @property
    def is_condensed(self) -> bool:
        return self.is_condensed_raw is True

    @staticmethod
    @override
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "StandardLengthType":
        kwargs = dataclass_fields_asdict(DiagCodedType.from_et(et_element, context))

        bit_length = int(odxrequire(et_element.findtext("BIT-LENGTH")))
        bit_mask = read_hex_binary(et_element.find("BIT-MASK"))
        is_condensed_raw = odxstr_to_bool(et_element.get("IS-CONDENSED"))

        return StandardLengthType(
            bit_length=bit_length, bit_mask=bit_mask, is_condensed_raw=is_condensed_raw, **kwargs)

    def __post_init__(self) -> None:
        if self.bit_mask is not None:
            maskable_types = (DataType.A_UINT32, DataType.A_INT32, DataType.A_BYTEFIELD)
            odxassert(
                self.base_data_type in maskable_types,
                'Can not apply a bit_mask on a value of type {self.base_data_type}',
            )

    def __get_used_mask(self, internal_value: AtomicOdxType) -> bytes | None:
        """Returns a byte field where all bits that are used by the
        DiagCoded type are set and all unused ones are not set.

        If `None` is returned, all bits are used.
        """
        if self.bit_mask is None:
            return None

        endianness: Literal["little", "big"] = "big"
        if not self.is_highlow_byte_order and self.base_data_type in [
                DataType.A_INT32, DataType.A_UINT32, DataType.A_FLOAT32, DataType.A_FLOAT64
        ]:
            # TODO (?): Technically, little endian A_UNICODE2STRING
            # objects require a byte swap for each 16 bit letter, and
            # thus also for the mask. I somehow doubt that this has
            # been anticipated by the standard, though...
            endianness = "little"

        if self.is_condensed:
            # if a condensed bitmask is specified, the number of bits
            # set to one in the bit mask are used in the PDU
            bit_sz = self.bit_mask.bit_count()
            used_mask = (1 << bit_sz) - 1

            return used_mask.to_bytes((bit_sz + 7) // 8, endianness)

        sz: int
        if isinstance(internal_value, BytesTypes):
            sz = len(bytes(internal_value))
        else:
            sz = (odxrequire(self.get_static_bit_length()) + 7) // 8

        max_value = (1 << (sz * 8)) - 1
        used_mask = self.bit_mask & max_value

        return used_mask.to_bytes(sz, endianness)

    def __apply_mask(self, internal_value: AtomicOdxType) -> AtomicOdxType:
        if self.bit_mask is None:
            return internal_value

        if self.is_condensed:
            int_value: int
            if isinstance(internal_value, BytesTypes):
                int_value = int.from_bytes(internal_value, 'big')
            elif isinstance(internal_value, int):
                int_value = internal_value
            else:
                odxraise("bit masks can only be specified for integers and byte fields")
                return None

            result = 0
            mask_bit = 0
            result_bit = 0

            while self.bit_mask >= (1 << mask_bit):
                if self.bit_mask & (1 << mask_bit):
                    result |= ((int_value & (1 << mask_bit)) >> mask_bit) << result_bit
                    result_bit += 1

                mask_bit += 1

            if isinstance(internal_value, BytesTypes):
                return result.to_bytes(len(bytes(internal_value)), 'big')

            return result

        if isinstance(internal_value, int):
            return internal_value & self.bit_mask
        if isinstance(internal_value, BytesTypes):
            int_value = int.from_bytes(internal_value, 'big')
            int_value &= self.bit_mask
            return int_value.to_bytes(len(bytes(internal_value)), 'big')

        odxraise(f'Can not apply a bit_mask on a value of type {type(internal_value)}')
        return internal_value

    def __unapply_mask(self, raw_value: AtomicOdxType) -> AtomicOdxType:
        if self.bit_mask is None:
            return raw_value
        if self.is_condensed:
            int_value: int
            if isinstance(raw_value, BytesTypes):
                int_value = int.from_bytes(raw_value, 'big')
            elif isinstance(raw_value, int):
                int_value = raw_value
            else:
                odxraise("bit masks can only be specified for integers and byte fields")
                return None

            result = 0
            mask_bit = 0
            input_bit = 0
            while self.bit_mask >= (1 << mask_bit):
                if self.bit_mask & (1 << mask_bit):
                    result |= ((int_value & (1 << input_bit)) >> input_bit) << mask_bit
                    input_bit += 1

                mask_bit += 1

            if isinstance(raw_value, BytesTypes):
                return result.to_bytes(len(bytes(raw_value)), 'big')

            return result
        if isinstance(raw_value, int):
            return raw_value & self.bit_mask
        if isinstance(raw_value, BytesTypes):
            int_value = int.from_bytes(raw_value, 'big')
            int_value &= self.bit_mask
            return int_value.to_bytes(len(bytes(raw_value)), 'big')

        odxraise(f'Can not apply a bit_mask on a value of type {type(raw_value)}')
        return raw_value

    def get_static_bit_length(self) -> int | None:
        if self.bit_mask is not None and self.is_condensed:
            return self.bit_mask.bit_count()

        return self.bit_length

    @override
    def encode_into_pdu(self, internal_value: AtomicOdxType, encode_state: EncodeState) -> None:
        encode_state.emplace_atomic_value(
            internal_value=self.__apply_mask(internal_value),
            used_mask=self.__get_used_mask(internal_value),
            bit_length=self.bit_length,
            base_data_type=self.base_data_type,
            base_type_encoding=self.base_type_encoding,
            is_highlow_byte_order=self.is_highlow_byte_order)

    @override
    def decode_from_pdu(self, decode_state: DecodeState) -> AtomicOdxType:
        raw_value = decode_state.extract_atomic_value(
            bit_length=self.bit_length,
            base_data_type=self.base_data_type,
            base_type_encoding=self.base_type_encoding,
            is_highlow_byte_order=self.is_highlow_byte_order)
        internal_value = self.__unapply_mask(raw_value)

        return internal_value



# SPDX-License-Identifier: MIT
from dataclasses import dataclass
from typing import Any
from xml.etree import ElementTree

from typing_extensions import override

from ..dataobjectproperty import DataObjectProperty
from ..decodestate import DecodeState
from ..encodestate import EncodeState
from ..exceptions import DecodeError, EncodeError, odxraise, odxrequire
from ..odxdoccontext import OdxDocContext
from ..odxlink import OdxLinkDatabase, OdxLinkId
from ..odxtypes import ParameterValue
from ..snrefcontext import SnRefContext
from ..utils import dataclass_fields_asdict
from .parameter import ParameterType
from .parameterwithdop import ParameterWithDOP


@dataclass(kw_only=True)
class PhysicalConstantParameter(ParameterWithDOP):
    physical_constant_value_raw: str

    @property
    @override
    def parameter_type(self) -> ParameterType:
        return "PHYS-CONST"

    @property
    @override
    def is_required(self) -> bool:
        return False

    @property
    @override
    def is_settable(self) -> bool:
        return False

    @property
    def physical_constant_value(self) -> ParameterValue:
        return self._physical_constant_value

    @staticmethod
    @override
    def from_et(et_element: ElementTree.Element,
                context: OdxDocContext) -> "PhysicalConstantParameter":

        kwargs = dataclass_fields_asdict(ParameterWithDOP.from_et(et_element, context))

        physical_constant_value_raw = odxrequire(et_element.findtext("PHYS-CONSTANT-VALUE"))

        return PhysicalConstantParameter(
            physical_constant_value_raw=physical_constant_value_raw, **kwargs)

    @override
    def _build_odxlinks(self) -> dict[OdxLinkId, Any]:
        return super()._build_odxlinks()

    @override
    def _resolve_odxlinks(self, odxlinks: OdxLinkDatabase) -> None:
        super()._resolve_odxlinks(odxlinks)

    @override
    def _resolve_snrefs(self, context: SnRefContext) -> None:
        super()._resolve_snrefs(context)

        dop = odxrequire(self.dop)
        if not isinstance(dop, DataObjectProperty):
            odxraise("The type of PHYS-CONST parameters must be a simple DOP")
            return
        base_data_type = dop.physical_type.base_data_type
        self._physical_constant_value = base_data_type.from_string(self.physical_constant_value_raw)

    @override
    def _encode_positioned_into_pdu(self, physical_value: ParameterValue | None,
                                    encode_state: EncodeState) -> None:
        if physical_value is not None and physical_value != self.physical_constant_value:
            odxraise(
                f"Value for constant parameter `{self.short_name}` name can "
                f"only be specified as {self.physical_constant_value!r} (is: {physical_value!r})",
                EncodeError)

        self.dop.encode_into_pdu(self.physical_constant_value, encode_state)

    @override
    def _decode_positioned_from_pdu(self, decode_state: DecodeState) -> ParameterValue:
        # Decode value
        phys_val = super()._decode_positioned_from_pdu(decode_state)

        # Check if decoded value matches expected value
        if phys_val != self.physical_constant_value:
            odxraise(
                f"Physical constant parameter does not match! "
                f"The parameter {self.short_name} expected physical value "
                f"{self.physical_constant_value!r} but got {phys_val!r} "
                f"at byte position {decode_state.cursor_byte_position} "
                f"in coded message {decode_state.coded_message.hex()}.", DecodeError)

        return phys_val



# SPDX-License-Identifier: MIT
import unittest
from xml.etree import ElementTree

from packaging.version import Version

from odxtools.odxdoccontext import OdxDocContext
from odxtools.odxlink import DocType, OdxDocFragment
from odxtools.odxtypes import DataType
from odxtools.parameters.createanyparameter import create_any_parameter_from_et
from odxtools.parameters.nrcconstparameter import NrcConstParameter

doc_frags = [OdxDocFragment("UnitTest", DocType.CONTAINER)]


class TestReadNrcParam(unittest.TestCase):

    def test_create_nrcconst_from_et(self) -> None:
        ODX = """
        <PARAM SEMANTIC="SUBFUNCTION-ID" xsi:type="NRC-CONST" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
            <SHORT-NAME>NR_identifier</SHORT-NAME>
            <LONG-NAME>Identifier for the negative response</LONG-NAME>
            <BYTE-POSITION>0</BYTE-POSITION>
            <CODED-VALUES>
                <CODED-VALUE>16</CODED-VALUE>
                <CODED-VALUE>10</CODED-VALUE>
            </CODED-VALUES>
            <DIAG-CODED-TYPE BASE-DATA-TYPE="A_UINT32" xsi:type="STANDARD-LENGTH-TYPE">
                <BIT-LENGTH>8</BIT-LENGTH>
            </DIAG-CODED-TYPE>
        </PARAM>
        """
        root = ElementTree.fromstring(ODX)
        param = create_any_parameter_from_et(root, OdxDocContext(Version("2.2.0"), doc_frags))

        self.assertIsInstance(param, NrcConstParameter)
        assert isinstance(param, NrcConstParameter)
        self.assertEqual("SUBFUNCTION-ID", param.semantic)
        self.assertEqual("NR_identifier", param.short_name)
        self.assertEqual("Identifier for the negative response", param.long_name)
        self.assertEqual([16, 10], param.coded_values)
        self.assertEqual(DataType.A_UINT32, param.diag_coded_type.base_data_type)
        self.assertEqual(8, param.diag_coded_type.get_static_bit_length())



# SPDX-License-Identifier: MIT
from dataclasses import dataclass
from xml.etree import ElementTree

from .matchingparameter import MatchingParameter
from .odxdoccontext import OdxDocContext
from .odxtypes import odxstr_to_bool
from .utils import dataclass_fields_asdict


@dataclass(kw_only=True)
class MatchingBaseVariantParameter(MatchingParameter):
    """A description of a parameter used for base variant matching.

    This is very similar to `MatchingParameter` for ECU variant
    matching, but `MatchingBaseVariantParameter` features the
    additional subtag `USE-PHYSICAL-ADDRESSING`.
    """

    use_physical_addressing_raw: bool | None = None

    @property
    def use_physical_addressing(self) -> bool:
        return self.use_physical_addressing_raw in [None, True]

    @staticmethod
    def from_et(et_element: ElementTree.Element,
                context: OdxDocContext) -> "MatchingBaseVariantParameter":

        kwargs = dataclass_fields_asdict(MatchingParameter.from_et(et_element, context))

        use_physical_addressing_raw = odxstr_to_bool(et_element.findtext("USE-PHYSICAL-ADDRESSING"))

        return MatchingBaseVariantParameter(
            use_physical_addressing_raw=use_physical_addressing_raw,
            **kwargs,
        )



# SPDX-License-Identifier: MIT
import pickle
import unittest
from dataclasses import dataclass

import odxtools_porsche
import odxtools_porsche.exceptions
from odxtools.exceptions import OdxError
from odxtools.loadfile import load_pdx_file
from odxtools.nameditemlist import NamedItemList
from odxtools.odxlink import OdxLinkRef

odxdb = load_pdx_file("./examples/somersault.pdx")

# use the diag layer container's document fragments as the default for
# resolving references
container_doc_frags = odxdb.diag_layer_containers.somersault.odx_id.doc_fragments


class TestStrictMode(unittest.TestCase):

    def test_strict_mode(self) -> None:
        # by default, we must be in strict mode
        self.assertTrue(odxtools.exceptions.strict_mode)

        # in strict mode, odxraise() must raise OdxError ...
        with self.assertRaises(OdxError):
            odxtools.exceptions.odxraise()
        # ... odxassert() must raise OdxError for failed
        # assertations ...
        odxtools.exceptions.odxassert(True)  # nothing happens
        with self.assertRaises(OdxError):
            odxtools.exceptions.odxassert(False)
        # ... and odxrequire() raises OdxError if passed `None` and
        # passes through everything else.
        with self.assertRaises(OdxError):
            odxtools.exceptions.odxrequire(None)
        self.assertEqual(odxtools.exceptions.odxrequire(123), 123)

        # change to non-strict mode
        odxtools.exceptions.strict_mode = False

        # in non-strict mode none of the above functions raises anything
        odxtools.exceptions.odxraise()
        odxtools.exceptions.odxassert(True)  # nothing happens
        odxtools.exceptions.odxassert(False)
        odxtools.exceptions.odxrequire(None)
        self.assertEqual(odxtools.exceptions.odxrequire(123), 123)

        # go back to strict mode
        odxtools.exceptions.strict_mode = True


class TestDataObjectProperty(unittest.TestCase):

    def test_bit_length(self) -> None:
        self.dop = odxdb.odxlinks.resolve(
            OdxLinkRef("somersault.DOP.num_flips", container_doc_frags))
        self.assertEqual(self.dop.get_static_bit_length(), 8)

    def test_convert_physical_to_internal(self) -> None:
        self.dop = odxdb.odxlinks.resolve(OdxLinkRef("somersault.DOP.boolean", container_doc_frags))
        self.assertEqual(self.dop.compu_method.convert_physical_to_internal("false"), 0)
        self.assertEqual(self.dop.compu_method.convert_physical_to_internal("true"), 1)


class TestComposeUDS(unittest.TestCase):

    def test_encode_with_coded_const(self) -> None:
        request = odxdb.odxlinks.resolve(
            OdxLinkRef("somersault.RQ.tester_present", container_doc_frags))
        self.assertEqual(bytes(request.encode()), 0x3E00.to_bytes(2, "big"))

    def test_encode_with_texttable(self) -> None:
        request = odxdb.odxlinks.resolve(
            OdxLinkRef("somersault.RQ.set_operation_params", container_doc_frags))
        self.assertEqual(
            bytes(request.encode(**{"use_fire_ring": "true"})), 0xBD01.to_bytes(2, "big"))
        self.assertEqual(bytes(request.encode(use_fire_ring="false")), 0xBD00.to_bytes(2, "big"))

    def test_encode_response_with_matching_request_param_and_structure(self) -> None:
        request = odxdb.odxlinks.resolve(
            OdxLinkRef("somersault.RQ.do_forward_flips", container_doc_frags))
        response = odxdb.odxlinks.resolve(
            OdxLinkRef("somersault.PR.happy_forward", container_doc_frags))

        coded_request = request.encode(forward_soberness_check=0x12, num_flips=12)
        self.assertEqual(bytes(coded_request), bytes.fromhex("ba120c"))
        coded_response = response.encode(yeha_level=3, coded_request=coded_request)
        self.assertEqual(bytes(coded_response), bytes.fromhex("faba03"))


@dataclass
class X:
    short_name: str
    value: int


class TestNamedItemList(unittest.TestCase):

    def test_NamedItemList(self) -> None:
        foo = NamedItemList([X("hello", 0), X("world", 1)])

        pickled_foo = pickle.dumps(foo)
        unpickled_foo = pickle.loads(pickled_foo)

        # Test with the original object
        self._test_NamedItemList_functionality(foo)
        # Test with the pickled object
        self._test_NamedItemList_functionality(unpickled_foo)

    def _test_NamedItemList_functionality(self, foo: NamedItemList[X]) -> None:
        self.assertEqual(foo.hello, X("hello", 0))
        self.assertEqual(foo[0], X("hello", 0))
        self.assertEqual(foo[1], X("world", 1))
        self.assertEqual(foo[:1], [X("hello", 0)])
        self.assertEqual(foo[-1:], [X("world", 1)])
        with self.assertRaises(IndexError):
            foo[2]
        self.assertEqual(foo["hello"], X("hello", 0))
        self.assertEqual(foo["world"], X("world", 1))
        self.assertEqual(foo.hello, X("hello", 0))
        self.assertEqual(foo.world, X("world", 1))

        foo.append(X("hello", 2))
        self.assertEqual(foo[2], X("hello", 2))
        self.assertEqual(foo["hello"], X("hello", 0))
        self.assertEqual(foo["hello_2"], X("hello", 2))
        self.assertEqual(foo.hello, X("hello", 0))
        self.assertEqual(foo.hello_2, X("hello", 2))

        # try to append an item that cannot be mapped to a name
        with self.assertRaises(OdxError):
            foo.append((0, 3))  # type: ignore[arg-type]

        # add a keyword identifier
        foo.append(X("as", 3))
        self.assertEqual(foo[3], X("as", 3))
        self.assertEqual(foo["_as"], X("as", 3))
        self.assertEqual(foo._as, X("as", 3))

        # add an object which's name conflicts with a method of the class
        foo.append(X("sort", 4))
        self.assertEqual(foo[4], X("sort", 4))
        self.assertEqual(foo["sort_2"], X("sort", 4))
        self.assertEqual(foo.sort_2, X("sort", 4))

        # test the get() function
        self.assertEqual(foo.get(0), X("hello", 0))
        self.assertEqual(foo.get(1234), None)
        self.assertEqual(foo.get(1234, X("foo", 5678)), X("foo", 5678))

        self.assertEqual(foo.get("hello"), X("hello", 0))
        self.assertEqual(foo.get("dunno"), None)
        self.assertEqual(foo.get("dunno", X("woho!", 0)), X("woho!", 0))

        # ensure that we can iterate over NamedItemList objects
        for x in foo:
            self.assertTrue(x in foo)

        # make sure that the keys(), values() and items() exist
        self.assertEqual(set(foo.keys()), {"hello", "world", "hello_2", "_as", "sort_2"})
        self.assertEqual(len(foo.items()), len(foo))
        self.assertEqual(len(foo.values()), len(foo))

        # ensure that mypy accepts NamedItemList objecs where List
        # objects are expected
        def bar(x: list[X]) -> None:
            pass

        bar(foo)


class TestNavigation(unittest.TestCase):

    def test_find_ecu_by_name(self) -> None:
        with self.assertRaises(KeyError):
            odxdb.ecus["somersault_crazy"]
        with self.assertRaises(IndexError):
            odxdb.ecus[len(odxdb.ecus) + 10]

        ecu = odxdb.ecus.get("somersault_crazy")
        self.assertEqual(ecu, None)

        ecu = odxdb.ecus.get(len(odxdb.ecus) + 10)
        self.assertEqual(ecu, None)

        # make sure that NamedItemLists support slicing
        ecus = odxdb.ecus[-2:]
        self.assertEqual(len(ecus), 2)
        self.assertEqual(ecus[0].odx_id.local_id, "somersault_lazy")
        self.assertEqual(ecus[1].odx_id.local_id, "somersault_assiduous")

        ecu = odxdb.ecus["somersault_lazy"]
        self.assertEqual(ecu.odx_id.local_id, "somersault_lazy")

        ecu = odxdb.ecus[0]
        self.assertEqual(ecu.odx_id.local_id, "somersault_lazy")

    def test_find_service_by_name(self) -> None:
        ecu = odxdb.ecus["somersault_lazy"]

        service_names = [s.short_name for s in ecu.services]
        self.assertIn("session_start", service_names)
        self.assertIn("session_stop", service_names)
        self.assertIn("tester_present", service_names)
        self.assertIn("do_forward_flips", service_names)
        self.assertNotIn("do_backward_flips", service_names)
        self.assertIn("report_status", service_names)

        service = ecu.services.session_start
        self.assertEqual(service.odx_id.local_id, "somersault.service.session_start")
        self.assertEqual(service.semantic, "SESSION")

    def test_udsbinner(self) -> None:
        ecu = odxdb.ecus.somersault_lazy

        service_groups = ecu.service_groups

        self.assertEqual(len(service_groups._service_groups), 5)
        self.assertEqual([s.short_name for s in service_groups[0x10]],
                         ["session_start", "session_stop"])
        self.assertEqual(service_groups[0x10].session_start.short_name, "session_start")
        self.assertEqual([s.short_name for s in service_groups[0xba]], ["do_forward_flips"])
        self.assertTrue(0x10 in service_groups)
        self.assertTrue(0x42 not in service_groups)
        self.assertEqual(service_groups[0x42], NamedItemList())


if __name__ == "__main__":
    unittest.main()



# SPDX-License-Identifier: MIT
import dataclasses
import re
from typing import TYPE_CHECKING, Any, Optional
from xml.etree import ElementTree

from .exceptions import odxraise

if TYPE_CHECKING:
    from .database import Database
    from .diaglayers.diaglayer import DiagLayer
    from .snrefcontext import SnRefContext


def read_hex_binary(et_element: ElementTree.Element | None) -> int | None:
    """Convert the contents of an xsd:hexBinary to an integer
    """
    if et_element is None:
        return None

    if (bytes_str := et_element.text) is None:
        # tag exists but is immediately terminated ("<FOO />"). we
        # treat this like an empty string.
        return 0

    # The XSD uses the type xsd:hexBinary and xsd:hexBinary allows for
    # leading/trailing whitespace and empty strings whilst `int(x,
    # 16)` raises an exception if one of these things happen.
    bytes_str = bytes_str.strip()
    if len(bytes_str) == 0:
        return 0

    try:
        return int(bytes_str, 16)
    except Exception as e:
        odxraise(f"Caught exception while parsing hex string `{bytes_str}`"
                 f" of {et_element.tag}: {e}")
        return None


def retarget_snrefs(database: "Database",
                    diag_layer: "DiagLayer",
                    context: Optional["SnRefContext"] = None) -> None:
    """Re-resolve the short name references reachable by a
    DiagLayer to this DiagLayer

    This implies that after the SNREFs have been retargeted, accessing
    the resolved objects via a different diagnostic layer might not be
    correct. E.g.: If the ECU variants "V1" and "V2" are derived from
    the base variant "BV", BV defines a short name reference to a data
    object property called "Foo" and V1 and V2 both define a "Foo"
    DOP, the reference in the base variant to Foo ought to be resolved
    differently depending on whether it is accessed via V1 or
    V2. Since odxtools resolves all references ahead of time, a fixed
    variant has to be chosen. This method allows to switch the variant
    to another one.

    """
    from .snrefcontext import SnRefContext

    if context is None:
        context = SnRefContext()

    if context.database is None:
        context.database = database

    if context.diag_layer is None:
        context.diag_layer = diag_layer

    # retarget the objects "owned" by the layer itself
    diag_layer._resolve_snrefs(context)

    # retarget all parents of the current layer (if any)
    if (parent_refs := getattr(diag_layer, "parent_refs", None)) is not None:
        for pr in parent_refs:
            retarget_snrefs(database, pr.layer, context)


def dataclass_fields_asdict(obj: Any) -> dict[str, Any]:
    """Extract all attributes from a dataclass object that are fields.

    This is a non-recursive version of `dataclasses.asdict()`. Its
    purpose is to make hierarchies of dataclasses possible while
    initializing the base class using common code.

    """
    return {x.name: getattr(obj, x.name) for x in dataclasses.fields(obj)}


# ISO 22901 section 7.1.1
_short_name_pattern = re.compile("[a-zA-Z0-9_]+")
# ISO 22901 section 7.3.13.3
_short_name_path_pattern = re.compile("[a-zA-Z0-9_]+(.[a-zA-Z0-9_]+)*")


def is_short_name(test_val: str) -> bool:
    """Returns true iff the test_val string is a ODX short name.

    See also: ISO 22901 section 7.1.1
    """
    return _short_name_pattern.fullmatch(test_val) is not None


def is_short_name_path(test_val: str) -> bool:
    """Returns true iff the test_val string is a ODX short name path.

    See also: ISO 22901 section 7.3.13.3
    """
    return _short_name_path_pattern.fullmatch(test_val) is not None



# SPDX-License-Identifier: MIT
from dataclasses import dataclass, field
from typing import Any
from xml.etree import ElementTree

from .diagservice import DiagService
from .element import IdentifiableElement
from .exceptions import odxassert, odxraise, odxrequire
from .nameditemlist import NamedItemList
from .odxdoccontext import OdxDocContext
from .odxlink import OdxLinkDatabase, OdxLinkId, resolve_snref
from .parameters.parameter import Parameter
from .snrefcontext import SnRefContext
from .utils import dataclass_fields_asdict


@dataclass(kw_only=True)
class SubComponentParamConnector(IdentifiableElement):
    diag_comm_snref: str

    # TODO: we currently only support SNREFs, not SNPATHREFs
    out_param_if_refs: list[str] = field(default_factory=list)
    in_param_if_refs: list[str] = field(default_factory=list)

    @property
    def service(self) -> DiagService:
        return self._service

    @property
    def out_param_ifs(self) -> NamedItemList[Parameter]:
        return self._out_param_ifs

    @property
    def in_param_ifs(self) -> NamedItemList[Parameter]:
        return self._in_param_ifs

    @staticmethod
    def from_et(et_element: ElementTree.Element,
                context: OdxDocContext) -> "SubComponentParamConnector":
        kwargs = dataclass_fields_asdict(IdentifiableElement.from_et(et_element, context))

        diag_comm_snref = odxrequire(
            odxrequire(et_element.find("DIAG-COMM-SNREF")).get("SHORT-NAME"))

        out_param_if_refs = []
        for elem in et_element.find("OUT-PARAM-IF-REFS") or []:
            if elem.tag != "OUT-PARAM-IF-SNREF":
                odxraise("Currently, only SNREFS are supported for OUT-PARAM-IF-REFS")
                continue
            else:
                odxassert(elem.tag == "OUT-PARAM-IF-SNREF")
                out_param_if_refs.append(odxrequire(elem.attrib.get("SHORT-NAME")))

        in_param_if_refs = []
        for elem in et_element.find("IN-PARAM-IF-REFS") or []:
            if elem.tag != "IN-PARAM-IF-SNREF":
                odxraise("Currently, only SNREFS are supported for IN-PARAM-IF-REFS")
                continue
            else:
                odxassert(elem.tag == "IN-PARAM-IF-SNREF")
                in_param_if_refs.append(odxrequire(elem.attrib.get("SHORT-NAME")))

        return SubComponentParamConnector(
            diag_comm_snref=diag_comm_snref,
            out_param_if_refs=out_param_if_refs,
            in_param_if_refs=in_param_if_refs,
            **kwargs)

    def _build_odxlinks(self) -> dict[OdxLinkId, Any]:
        return {}

    def _resolve_odxlinks(self, odxlinks: OdxLinkDatabase) -> None:
        pass

    def _resolve_snrefs(self, context: SnRefContext) -> None:
        service = resolve_snref(self.diag_comm_snref,
                                odxrequire(context.diag_layer).diag_comms, DiagService)
        self._service = service

        if self._service.request is not None:
            odxraise()
            return
        if not self._service.positive_responses:
            odxraise()
            return

        request = odxrequire(service.request)
        in_param_ifs = []
        for x in self.in_param_if_refs:
            in_param_ifs.append(resolve_snref(x, request.parameters, Parameter))

        # TODO: The output parameters are probably part of a response
        # (?). If so, they cannot be resolved ahead of time because
        # the service in question can have multiple responses
        # associated with it and each of these has its own set of
        # parameters. In the meantime, we simply use the first
        # positive response specified.
        response = service.positive_responses[0]
        out_param_ifs = []
        for x in self.out_param_if_refs:
            out_param_ifs.append(resolve_snref(x, response.parameters, Parameter))

        self._in_param_ifs = NamedItemList(in_param_ifs)
        self._out_param_ifs = NamedItemList(out_param_ifs)



# SPDX-License-Identifier: MIT
from dataclasses import dataclass
from typing import Any
from xml.etree import ElementTree

from typing_extensions import final, override

from ..decodestate import DecodeState
from ..encodestate import EncodeState
from ..exceptions import EncodeError, odxraise, odxrequire
from ..odxdoccontext import OdxDocContext
from ..odxlink import OdxLinkId
from ..odxtypes import ParameterValue
from ..utils import dataclass_fields_asdict
from .parameter import ParameterType
from .parameterwithdop import ParameterWithDOP


@dataclass(kw_only=True)
class LengthKeyParameter(ParameterWithDOP):
    """Length Keys specify the bit (!) length of another parameter.

    The other parameter references the length key parameter using a ParamLengthInfoType as .diag_coded_type.

    LengthKeyParameters are decoded the same as ValueParameters.
    The main difference to ValueParameters is that a LengthKeyParameter has an `.odx_id` attribute
    and its DOP must be a simple DOP with PHYSICAL-TYPE/BASE-DATA-TYPE=DataType.A_UINT32.
    """

    odx_id: OdxLinkId

    @property
    @override
    def parameter_type(self) -> ParameterType:
        return "LENGTH-KEY"

    @property
    @override
    def is_required(self) -> bool:
        return False

    @property
    @override
    def is_settable(self) -> bool:
        # length keys can be explicitly set, but they do not need to
        # be because they can be implicitly determined by the length
        # of the corresponding field
        return True

    @staticmethod
    @override
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "LengthKeyParameter":

        kwargs = dataclass_fields_asdict(ParameterWithDOP.from_et(et_element, context))

        odx_id = odxrequire(OdxLinkId.from_et(et_element, context))

        return LengthKeyParameter(odx_id=odx_id, **kwargs)

    @override
    def _build_odxlinks(self) -> dict[OdxLinkId, Any]:
        result = super()._build_odxlinks()

        result[self.odx_id] = self

        return result

    @override
    @final
    def _encode_positioned_into_pdu(self, physical_value: ParameterValue | None,
                                    encode_state: EncodeState) -> None:
        # if you get this exception, you ought to use
        # `.encode_placeholder_into_pdu()` followed by (after the
        # value of the length key has been determined)
        # `.encode_value_into_pdu()`.
        raise RuntimeError("_encode_positioned_into_pdu() cannot be called for length keys.")

    def encode_placeholder_into_pdu(self, physical_value: ParameterValue | None,
                                    encode_state: EncodeState) -> None:

        if physical_value is not None:
            if not self.dop.is_valid_physical_value(physical_value):
                odxraise(f"Invalid explicitly specified physical value '{physical_value!r}' "
                         f"for length key '{self.short_name}'.")

            lkv = encode_state.length_keys.get(self.short_name)
            if lkv is not None and lkv != physical_value:
                odxraise(f"Got conflicting values for length key {self.short_name}: "
                         f"{lkv} and {physical_value!r}")

            if not isinstance(physical_value, int):
                odxraise(
                    f"Value of length key {self.short_name} is of type {type(physical_value).__name__} "
                    f"instead of int")

            encode_state.length_keys[self.short_name] = physical_value

        pos = encode_state.cursor_byte_position
        if self.byte_position is not None:
            pos = encode_state.origin_byte_position + self.byte_position
        encode_state.key_pos[self.short_name] = pos
        encode_state.cursor_byte_position = pos
        encode_state.cursor_bit_position = self.bit_position or 0

        # emplace a value of zero into the encode state, but pretend the bits not to be used
        n = odxrequire(self.dop.get_static_bit_length()) + encode_state.cursor_bit_position
        tmp_val = b'\x00' * ((n + 7) // 8)
        encode_state.cursor_bit_position = 0
        encode_state.emplace_bytes(tmp_val, obj_used_mask=tmp_val)

    def encode_value_into_pdu(self, encode_state: EncodeState) -> None:

        if self.short_name not in encode_state.length_keys:
            odxraise(
                f"Length key {self.short_name} has not been defined before "
                f"it is required.", EncodeError)
            return
        else:
            physical_value = encode_state.length_keys[self.short_name]

        encode_state.cursor_byte_position = encode_state.key_pos[self.short_name]
        encode_state.cursor_bit_position = self.bit_position or 0

        self.dop.encode_into_pdu(encode_state=encode_state, physical_value=physical_value)

    @override
    def _decode_positioned_from_pdu(self, decode_state: DecodeState) -> ParameterValue:
        phys_val = super()._decode_positioned_from_pdu(decode_state)

        if not isinstance(phys_val, int):
            odxraise(f"The pysical type of length keys must be an integer, "
                     f"(is {type(phys_val).__name__})")
        decode_state.length_keys[self.short_name] = phys_val

        return phys_val



# SPDX-License-Identifier: MIT
import unittest
from io import StringIO
from unittest.mock import patch

from packaging.version import Version

from odxtools.description import Description
from odxtools.exceptions import OdxError, odxrequire
from odxtools.loadfile import load_pdx_file
from odxtools.parameters.nrcconstparameter import NrcConstParameter
from odxtools.parameters.valueparameter import ValueParameter
from odxtools.utils import retarget_snrefs

odxdb = load_pdx_file("./examples/somersault.pdx")


class TestDatabase(unittest.TestCase):

    def test_db_structure(self) -> None:
        self.assertEqual(odxdb.model_version, Version("2.2.0"))
        self.assertEqual([x.short_name for x in odxdb.diag_layer_containers], ["somersault"])

        self.assertEqual(
            {x.short_name
             for x in odxdb.diag_layers},
            {"somersault", "somersault_assiduous", "somersault_lazy", "somersault_protocol"},
        )

        self.assertEqual([x.short_name for x in odxdb.ecus],
                         ["somersault_lazy", "somersault_assiduous"])

    def test_admin_data(self) -> None:
        dlc = odxdb.diag_layer_containers.somersault

        self.assertTrue(dlc.admin_data is not None)
        ad = odxrequire(dlc.admin_data)

        self.assertEqual(ad.language, "en-US")

        self.assertTrue(ad.company_doc_infos is not None)
        cdi = ad.company_doc_infos

        self.assertEqual(len(cdi), 1)
        self.assertEqual(cdi[0].company_data_ref.ref_id, "CD.Suncus")
        self.assertTrue(cdi[0].company_data is not None)
        self.assertEqual(odxrequire(cdi[0].team_member_ref).ref_id, "TM.Doggy")
        self.assertTrue(cdi[0].team_member is not None)
        self.assertEqual(cdi[0].doc_label, "A really meaningful label")

        self.assertTrue(ad.doc_revisions is not None)
        dr = ad.doc_revisions[0]

        self.assertEqual(len(dr.modifications), 2)
        mod = dr.modifications[0]

        self.assertEqual(mod.change, "add somersault ECU")
        self.assertEqual(mod.reason, "we needed a new artist")

        self.assertEqual(dr.revision_label, "1.0")
        self.assertEqual(odxrequire(dr.team_member_ref).ref_id, "TM.Doggy")
        self.assertTrue(dr.team_member is not None)
        self.assertEqual(dr.tool, "odxtools 0.0.1")

    def test_company_datas(self) -> None:
        dlc = odxdb.diag_layer_containers.somersault

        self.assertTrue(dlc.company_datas is not None)
        cds = dlc.company_datas

        self.assertEqual([x.short_name for x in cds], ["Suncus", "ACME_Corporation"])

        cd = cds.Suncus
        self.assertEqual(cd.odx_id.local_id, "CD.Suncus")
        self.assertEqual(cd.short_name, "Suncus")
        self.assertEqual(cd.long_name, "Circus of the sun")
        self.assertEqual(cd.description,
                         Description.from_string("<p>Prestigious group of performers</p>"))
        self.assertEqual(cd.roles, ["circus", "gym"])

        self.assertEqual([x.short_name for x in cd.team_members], ["Doggy", "Horsey"])

        doggy = cd.team_members.Doggy
        self.assertEqual(doggy.odx_id.local_id, "TM.Doggy")
        self.assertEqual(doggy.short_name, "Doggy")
        self.assertEqual(doggy.long_name, "Doggy the dog")
        self.assertEqual(doggy.description,
                         Description.from_string("<p>Dog is man's best friend</p>"))
        self.assertEqual(doggy.roles, ["gymnast", "tracker"])
        self.assertEqual(doggy.department, "sniffers")
        self.assertEqual(doggy.address, "Some road")
        self.assertEqual(doggy.zipcode, "12345")
        self.assertEqual(doggy.city, "New Dogsville")
        self.assertEqual(doggy.phone, "+0 1234/5678-9")
        self.assertEqual(doggy.fax, "+0 1234/5678-0")
        self.assertEqual(doggy.email, "info@suncus.com")

        self.assertTrue(cd.company_specific_info is not None)
        self.assertTrue(odxrequire(cd.company_specific_info).related_docs is not None)

        rd = odxrequire(cd.company_specific_info).related_docs[0]

        self.assertEqual(rd.description, Description.from_string("<p>We are the best!</p>"))
        self.assertTrue(rd.xdoc is not None)

        xdoc = odxrequire(rd.xdoc)
        self.assertEqual(xdoc.short_name, "best")
        self.assertEqual(xdoc.long_name, "suncus is the best")
        self.assertEqual(xdoc.description, Description.from_string("<p>great propaganda...</p>"))
        self.assertEqual(xdoc.number, "1")
        self.assertEqual(xdoc.state, "published")
        self.assertEqual(xdoc.date, "2015-01-15T20:15:20+05:00")
        self.assertEqual(xdoc.publisher, "Suncus Publishing")
        self.assertEqual(xdoc.url, "https://suncus-is-the-best.com")
        self.assertEqual(xdoc.position, "first!")

    def test_somersault_lazy(self) -> None:
        # TODO: this test is far from exhaustive
        ecu = odxdb.ecus.somersault_lazy

        self.assertEqual(
            {x.short_name
             for x in ecu.diag_comms},
            {
                "compulsory_program",
                "do_forward_flips",
                "report_status",
                "session_start",
                "session_stop",
                "tester_present",
                "schroedinger",
            },
        )

        service = ecu.services.do_forward_flips
        self.assertEqual(
            [x.short_name for x in odxrequire(service.request).parameters],
            ["sid", "forward_soberness_check", "num_flips"],
        )
        self.assertEqual(
            [x.short_name for x in odxrequire(service.request).required_parameters],
            ["forward_soberness_check", "num_flips"],
        )
        self.assertEqual(odxrequire(service.request).get_static_bit_length(), 24)

        self.assertEqual([x.short_name for x in service.positive_responses], ["grudging_forward"])
        self.assertEqual([x.short_name for x in service.negative_responses], ["flips_not_done"])

        pr = service.positive_responses.grudging_forward
        self.assertEqual([x.short_name for x in pr.parameters],
                         ["sid", "num_flips_done", "sault_time"])
        self.assertEqual([x.short_name for x in pr.required_parameters], [])
        self.assertEqual([x.short_name for x in pr.free_parameters], ["sault_time"])
        self.assertEqual(pr.get_static_bit_length(), 24)

        nr = service.negative_responses.flips_not_done
        self.assertEqual(
            [x.short_name for x in nr.parameters],
            ["sid", "rq_sid", "reason", "flips_successfully_done"],
        )
        self.assertEqual(nr.get_static_bit_length(), 32)

        nrc_const = nr.parameters.reason
        self.assertEqual(nrc_const.parameter_type, "NRC-CONST")
        assert isinstance(nrc_const, NrcConstParameter)
        self.assertEqual(nrc_const.coded_values, [0, 1, 2])


class TestEnDecode(unittest.TestCase):

    def test_encode_specify_unknown_param(self) -> None:
        ecu = odxdb.ecus.somersault_lazy
        service = ecu.services.do_forward_flips
        request = odxrequire(service.request)
        with self.assertRaises(OdxError) as eo:
            request.encode(forward_soberness_check=0x12, num_flips=5, grass_level="what grass?")

        self.assertEqual(
            str(eo.exception),
            "Value for unknown parameter 'grass_level' specified for composite codec object do_forward_flips"
        )

    def test_decode_request(self) -> None:
        messages = odxdb.ecus.somersault_assiduous.decode(bytes([0x03, 0x45]))
        self.assertTrue(len(messages) == 1)
        m = messages[0]
        self.assertEqual(m.coded_message, bytes([0x03, 0x45]))
        self.assertEqual(m.service, odxdb.ecus.somersault_assiduous.services.headstand)
        self.assertEqual(m.coding_object,
                         odxdb.ecus.somersault_assiduous.services.headstand.request)
        self.assertEqual(m.param_dict, {"sid": 0x03, "duration": 0x45})

    def test_decode_global_negative_response(self) -> None:
        ecu = odxdb.ecus.somersault_assiduous
        coded_request = bytes([0x03, 0x45])

        self.assertEqual(len(ecu.global_negative_responses), 1)

        gnr = ecu.global_negative_responses.too_hot
        coded_response = gnr.encode(coded_request=coded_request, temperature=35)

        decoded = ecu.decode(coded_response)
        # the global negative response for the somersault ECUs does
        # not include any matching-request parameter, so decode()
        # returns one possible instance per service
        self.assertEqual(len(decoded), len(ecu.services))

        # if we specify the request, the result should become unique
        decoded_resp = ecu.decode_response(coded_response, request=coded_request)
        self.assertEqual(len(decoded_resp), 1)

        # make sure that the global negative response was decoded
        # correctly
        resp = decoded_resp[0]
        self.assertEqual(resp.param_dict['temperature'], 35)

    def test_code_table_params(self) -> None:
        """en- and decode table parameters"""
        ecu = odxdb.ecus.somersault_assiduous
        pr = ecu.services.report_status.positive_responses.status_report

        # test the "no flips done yet" response
        resp_data = pr.encode(
            dizzyness_level=12,
            happiness_level=100,
            last_pos_response_key="none",  # <- name of the selected table row
            last_pos_response=("none", 123))
        self.assertEqual(resp_data.hex(), "620c64007b")

        decoded_resp_data = pr.decode(resp_data)
        assert isinstance(decoded_resp_data, dict)
        self.assertEqual(decoded_resp_data["dizzyness_level"], 12)
        self.assertEqual(decoded_resp_data["happiness_level"], 100)
        self.assertEqual(decoded_resp_data["last_pos_response_key"], "none")
        self.assertEqual(decoded_resp_data["last_pos_response"], ("none", 123))

        # test the "forward flips grudgingly done" response. we define
        # the table key implicitly by the 'last_pos_response'
        # table-struct parameter this time
        resp_data = pr.encode(
            coded_request=bytearray([123] * 15),
            dizzyness_level=42,
            happiness_level=92,
            last_pos_response=("forward_grudging", {
                "sault_time": 249
            }))
        self.assertEqual(resp_data.hex(), "622a5c03fa7bf9")

        # test decoding an object featuring a TableStruct parameter
        decoded_resp_data = pr.decode(resp_data)
        assert isinstance(decoded_resp_data, dict)
        self.assertEqual(decoded_resp_data["dizzyness_level"], 42)
        self.assertEqual(decoded_resp_data["happiness_level"], 92)
        self.assertEqual(decoded_resp_data["last_pos_response_key"], "forward_grudging")
        last_pos_response = decoded_resp_data["last_pos_response"]
        assert isinstance(last_pos_response, tuple)
        lpr_name, lpr_value = last_pos_response
        assert isinstance(lpr_name, str)
        assert isinstance(lpr_value, dict)
        self.assertEqual(lpr_name, "forward_grudging")
        self.assertEqual(set(lpr_value.keys()), {"sid", "num_flips_done", "sault_time"})
        # the num_flips_done parameter is a matching request parameter
        # for this response, so it produces a binary blob. possibly,
        # it should be changed to a ValueParameter...
        self.assertEqual(lpr_value["num_flips_done"], 123)
        self.assertEqual(lpr_value["sault_time"], 249)

        # test the "backward flips grudgingly done" response
        resp_data = pr.encode(
            dizzyness_level=75,
            happiness_level=3,
            last_pos_response=("backward_grudging", {
                'num_flips_done': 5,
                'grumpiness_level': 150
            }))
        self.assertEqual(resp_data.hex(), "624b030afb0596")

        decoded_resp_data = pr.decode(resp_data)
        assert isinstance(decoded_resp_data, dict)
        self.assertEqual(decoded_resp_data["dizzyness_level"], 75)
        self.assertEqual(decoded_resp_data["happiness_level"], 3)
        self.assertEqual(decoded_resp_data["last_pos_response_key"], "backward_grudging")
        last_pos_response = decoded_resp_data["last_pos_response"]
        self.assertTrue(isinstance(last_pos_response, tuple))
        assert isinstance(last_pos_response, tuple)
        self.assertEqual(len(last_pos_response), 2)
        self.assertEqual(last_pos_response[0], "backward_grudging")
        assert isinstance(last_pos_response[1], dict)
        self.assertEqual(
            set(last_pos_response[1].keys()), {"sid", "num_flips_done", "grumpiness_level"})
        self.assertEqual(last_pos_response[0], "backward_grudging")
        self.assertEqual(last_pos_response[1]["num_flips_done"], 5)
        self.assertEqual(last_pos_response[1]["grumpiness_level"], 150)

    def test_decode_inherited_request(self) -> None:
        raw_message = odxdb.ecus.somersault_assiduous.services.do_backward_flips(
            backward_soberness_check=0x21, num_flips=2)
        messages = odxdb.ecus.somersault_assiduous.decode(raw_message)
        self.assertTrue(len(messages) == 1)
        m = messages[0]
        self.assertEqual(m.coded_message, bytes([0xBB, 0x21, 0x02]))
        self.assertEqual(m.service, odxdb.ecus.somersault_assiduous.services.do_backward_flips)
        self.assertEqual(m.coding_object,
                         odxdb.ecus.somersault_assiduous.services.do_backward_flips.request)
        self.assertEqual(m.param_dict, {
            "sid": 0xBB,
            "backward_soberness_check": 0x21,
            "num_flips": 0x02
        })

    def test_free_param_info(self) -> None:
        ecu = odxdb.ecus.somersault_lazy
        service = ecu.services.do_forward_flips
        request = odxrequire(service.request)
        pos_response = service.positive_responses.grudging_forward
        neg_response = service.negative_responses.flips_not_done

        stdout = StringIO()
        with patch("sys.stdout", stdout):
            request.print_free_parameters_info()
            expected_output = "forward_soberness_check: uint\nnum_flips: uint\n"
            actual_output = stdout.getvalue()
            self.assertEqual(actual_output, expected_output)

        with patch("sys.stdout", stdout):
            pos_response.print_free_parameters_info()
            expected_output = "forward_soberness_check: uint\nnum_flips: uint\nsault_time: uint\n"
            actual_output = stdout.getvalue()
            self.assertEqual(actual_output, expected_output)

        with patch("sys.stdout", stdout):
            neg_response.print_free_parameters_info()
            expected_output = ("forward_soberness_check: uint\n"
                               "num_flips: uint\n"
                               "sault_time: uint\n"
                               "flips_successfully_done: uint\n")
            actual_output = stdout.getvalue()
            self.assertEqual(expected_output, actual_output)

    def test_decode_response(self) -> None:
        ecu = odxdb.ecus.somersault_lazy
        service = ecu.services.do_forward_flips
        raw_request_message = service(forward_soberness_check=0x12, num_flips=3)
        pos_response = service.positive_responses.grudging_forward

        raw_response_message = pos_response.encode(raw_request_message)

        messages = ecu.decode_response(raw_response_message, raw_request_message)
        self.assertTrue(
            len(messages) == 1,
            f"There should be only one service for 0x0145 but there are: {messages}",
        )
        m = messages[0]
        self.assertEqual(m.coded_message.hex(), "fa03ff")
        self.assertEqual(m.coding_object, pos_response)
        self.assertEqual(m.param_dict, {"sid": 0xFA, "num_flips_done": 0x03, "sault_time": 255})

    def test_retarget_snrefs(self) -> None:
        base_variant = odxdb.base_variants.somersault
        ecu_lazy = odxdb.ecu_variants.somersault_lazy
        ecu_assiduous = odxdb.ecu_variants.somersault_assiduous

        schroedinger_param_base = odxrequire(
            base_variant.services.schroedinger.request).parameters.schroedinger_param
        schroedinger_param_lazy = odxrequire(
            ecu_lazy.services.schroedinger.request).parameters.schroedinger_param
        schroedinger_param_assiduous = odxrequire(
            ecu_assiduous.services.schroedinger.request).parameters.schroedinger_param

        # the parameter object is the same regardless of how it has
        # been accessed.
        self.assertEqual(id(schroedinger_param_base), id(schroedinger_param_lazy))
        self.assertEqual(id(schroedinger_param_base), id(schroedinger_param_assiduous))

        # all Schroedinger parameters are the same, so let's skip the
        # qualifier from here on out
        schroedinger_param = schroedinger_param_base

        # by default, the DOP referenced by the Schroedinger parameter
        # is the one defined by the base variant
        assert isinstance(schroedinger_param, ValueParameter)
        self.assertEqual(schroedinger_param.dop.odx_id.local_id, "somersault.DOP.schroedinger_base")

        # if we retarget all short name references towards the lazy or
        # assiduous ECUs, the schroedinger parameter's DOP will change
        # accordingly (note that the parameter always references a DOP
        # called "schroedinger_dop" via SNREF but each diag layer
        # redefines this DOP)
        retarget_snrefs(odxdb, ecu_lazy)
        self.assertEqual(schroedinger_param.dop.odx_id.local_id, "somersault.DOP.schroedinger_lazy")

        retarget_snrefs(odxdb, ecu_assiduous)
        self.assertEqual(schroedinger_param.dop.odx_id.local_id,
                         "somersault.DOP.schroedinger_assiduous")

        # after the database is refreshed, it is back to the original state
        odxdb.refresh()
        self.assertEqual(schroedinger_param.dop.odx_id.local_id, "somersault.DOP.schroedinger_base")


class TestNavigation(unittest.TestCase):

    def test_finding_services(self) -> None:
        # Find base variant
        self.assertIsNotNone(odxdb.diag_layers.somersault.services.do_backward_flips)
        self.assertIsNotNone(odxdb.diag_layers.somersault.services.do_forward_flips)
        self.assertIsNotNone(odxdb.diag_layers.somersault.services.report_status)

        # Find ecu variant
        self.assertIsNotNone(odxdb.ecus.somersault_assiduous.services.headstand)
        # Inherited services
        self.assertIsNotNone(odxdb.ecus.somersault_assiduous.services.do_backward_flips)
        self.assertIsNotNone(odxdb.ecus.somersault_assiduous.services.do_forward_flips)
        self.assertIsNotNone(odxdb.ecus.somersault_assiduous.services.report_status)
        self.assertIsNotNone(odxdb.ecus.somersault_assiduous.single_ecu_jobs.compulsory_program)

        # The lazy ECU variant only inherits services but does not add any.
        self.assertIsNotNone(odxdb.ecus.somersault_lazy.services.do_forward_flips)
        self.assertIsNotNone(odxdb.ecus.somersault_lazy.services.report_status)

        # also, the lazy ECU does not do backward flips. (this is
        # reserved for swots...)
        with self.assertRaises(AttributeError):
            odxdb.ecus.somersault_lazy.services.do_backward_flips  # noqa: B018


if __name__ == "__main__":
    unittest.main()



# SPDX-License-Identifier: MIT
import warnings
from dataclasses import dataclass, field
from typing import TYPE_CHECKING

from .encoding import Encoding, get_string_encoding
from .exceptions import EncodeError, OdxWarning, odxassert, odxraise
from .odxtypes import AtomicOdxType, BytesTypes, DataType, ParameterValue

try:
    import bitstruct.c as bitstruct
except ImportError:
    import bitstruct

if TYPE_CHECKING:
    from .parameters.parameter import Parameter


@dataclass
class EncodeState:
    """Utility class to holding the state variables needed for encoding a message.
    """

    #: payload that has been constructed so far
    coded_message: bytearray = field(default_factory=bytearray)

    #: the bits of the payload that are used
    used_mask: bytearray = field(default_factory=bytearray)

    #: The absolute position in bytes from the beginning of the PDU to
    #: which relative positions refer to, e.g., the beginning of the
    #: structure.
    origin_byte_position: int = 0

    #: The absolute position in bytes from the beginning of the PDU
    #: where the next object ought to be placed into the PDU
    cursor_byte_position: int = 0

    #: The bit position [0-7] where the next object ought to be
    #: placed into the PDU
    cursor_bit_position: int = 0

    #: If encoding a response: request that triggered the response
    triggering_request: bytes | None = None

    #: Mapping from the short name of a length-key parameter to bit
    #: lengths (specified by LengthKeyParameter)
    length_keys: dict[str, int] = field(default_factory=dict)

    #: Mapping from the short name of a table-key parameter to the
    #: short name of the corresponding row of the table (specified by
    #: TableKeyParameter)
    table_keys: dict[str, str] = field(default_factory=dict)

    #: The cursor position where a given length- or table key is located
    #: in the PDU
    key_pos: dict[str, int] = field(default_factory=dict)

    #: Flag whether we are currently the last parameter of the PDU
    #: (needed for MinMaxLengthType, EndOfPduField, etc.)
    is_end_of_pdu: bool = True

    #: list of parameters that have been encoded so far. The journal
    #: is used by some types of parameters which depend on the values of
    #: other parameters; e.g., environment data description parameters
    journal: list[tuple["Parameter", ParameterValue | None]] = field(default_factory=list)

    #: If this is True, specifying unknown parameters for encoding
    #: will raise an OdxError exception in strict mode.
    allow_unknown_parameters = False

    def __post_init__(self) -> None:
        # if a coded message has been specified, but no used_mask, we
        # assume that all of the bits of the coded message are
        # currently used.
        if len(self.coded_message) > len(self.used_mask):
            self.used_mask += b'\xff' * (len(self.coded_message) - len(self.used_mask))
        if len(self.coded_message) < len(self.used_mask):
            odxraise(f"The specified bit mask 0x{self.used_mask.hex()} for used bits "
                     f"is not suitable for representing the coded_message "
                     f"0x{self.coded_message.hex()}")
            self.used_mask = self.used_mask[:len(self.coded_message)]

    def emplace_atomic_value(
        self,
        *,
        internal_value: AtomicOdxType,
        bit_length: int,
        base_data_type: DataType,
        base_type_encoding: Encoding | None,
        is_highlow_byte_order: bool,
        used_mask: bytes | None,
    ) -> None:
        """Convert the internal_value to bytes and emplace this into the PDU"""

        raw_value: AtomicOdxType

        # Deal with raw byte fields, ...
        if base_data_type == DataType.A_BYTEFIELD:
            if not isinstance(internal_value, BytesTypes):
                odxraise(f"{internal_value!r} is not a bytefield", EncodeError)
                return

            odxassert(
                base_type_encoding in (None, Encoding.NONE, Encoding.BCD_P, Encoding.BCD_UP),
                f"Illegal encoding '{base_type_encoding}' for A_BYTEFIELD")

            # note that we do not ensure that BCD-encoded byte fields
            # only represent "legal" values
            raw_value = bytes(internal_value)

            if 8 * len(raw_value) > bit_length:
                odxraise(
                    f"The value '{internal_value!r}' cannot be encoded using "
                    f"{bit_length} bits.", EncodeError)
                raw_value = raw_value[0:bit_length // 8]

        # ... string types, ...
        elif base_data_type in (DataType.A_UTF8STRING, DataType.A_ASCIISTRING,
                                DataType.A_UNICODE2STRING):
            if not isinstance(internal_value, str):
                odxraise(f"The internal value '{internal_value!r}' is not a string", EncodeError)
                internal_value = str(internal_value)

            str_encoding = get_string_encoding(base_data_type, base_type_encoding,
                                               is_highlow_byte_order)
            if str_encoding is not None:
                raw_value = internal_value.encode(str_encoding)
            else:
                raw_value = b""

            if 8 * len(raw_value) > bit_length:
                odxraise(
                    f"The value '{internal_value!r}' cannot be encoded using "
                    f"{bit_length} bits.", EncodeError)
                raw_value = raw_value[0:bit_length // 8]

        # ... signed integers, ...
        elif base_data_type == DataType.A_INT32:
            if not isinstance(internal_value, int):
                odxraise(
                    f"Internal value must be of integer type, not {type(internal_value).__name__}",
                    EncodeError)
                internal_value = int(internal_value)

            if base_type_encoding == Encoding.ONEC:
                # one-complement
                if internal_value >= 0:
                    raw_value = internal_value
                else:
                    mask = (1 << bit_length) - 1
                    raw_value = mask + internal_value
            elif base_type_encoding in (None, Encoding.TWOC):
                # two-complement
                if internal_value >= 0:
                    raw_value = internal_value
                else:
                    mask = (1 << bit_length) - 1
                    raw_value = mask + internal_value + 1
            elif base_type_encoding == Encoding.SM:
                # sign-magnitude
                if internal_value >= 0:
                    raw_value = internal_value
                else:
                    raw_value = (1 << (bit_length - 1)) + abs(internal_value)
            else:
                odxraise(
                    f"Illegal encoding ({base_type_encoding and base_type_encoding.value}) specified for "
                    f"{base_data_type.value}")

                if base_type_encoding == Encoding.BCD_P:
                    raw_value = self.__encode_bcd_p(abs(internal_value))
                elif base_type_encoding == Encoding.BCD_UP:
                    raw_value = self.__encode_bcd_up(abs(internal_value))
                else:
                    raw_value = internal_value

            if raw_value.bit_length() > bit_length:
                odxraise(
                    f"The value '{internal_value!r}' cannot be encoded using "
                    f"{bit_length} bits.", EncodeError)
                raw_value &= (1 << bit_length) - 1

        # ... unsigned integers, ...
        elif base_data_type == DataType.A_UINT32:
            if not isinstance(internal_value, int) or internal_value < 0:
                odxraise(f"Internal value must be a positive integer, not {internal_value!r}")
                internal_value = abs(int(internal_value))

            if base_type_encoding == Encoding.BCD_P:
                # packed BCD
                raw_value = self.__encode_bcd_p(internal_value)
            elif base_type_encoding == Encoding.BCD_UP:
                # unpacked BCD
                raw_value = self.__encode_bcd_up(internal_value)
            elif base_type_encoding in (None, Encoding.NONE):
                # no encoding
                raw_value = internal_value
            else:
                odxraise(f"Illegal encoding ({base_type_encoding}) specified for "
                         f"{base_data_type.value}")

                raw_value = internal_value

            if raw_value.bit_length() > bit_length:
                odxraise(
                    f"The value '{internal_value!r}' cannot be encoded using "
                    f"{bit_length} bits.", EncodeError)
                raw_value &= (1 << bit_length) - 1

        # ... and others (floating point values)
        else:
            odxassert(base_data_type in (DataType.A_FLOAT32, DataType.A_FLOAT64))
            odxassert(base_type_encoding in (None, Encoding.NONE))

            if base_data_type == DataType.A_FLOAT32 and bit_length != 32:
                odxraise(f"Illegal bit length for a float32 object ({bit_length})")
                bit_length = 32
            elif base_data_type == DataType.A_FLOAT64 and bit_length != 64:
                odxraise(f"Illegal bit length for a float64 object ({bit_length})")
                bit_length = 64

            raw_value = float(internal_value)  # type: ignore[arg-type]

        # If the bit length is zero, encode an empty value
        if bit_length == 0:
            self.emplace_bytes(b'')
            return

        format_char = base_data_type.bitstruct_format_letter
        padding = (8 - ((bit_length + self.cursor_bit_position) % 8)) % 8
        odxassert((0 <= padding and padding < 8 and
                   (padding + bit_length + self.cursor_bit_position) % 8 == 0),
                  f"Incorrect padding {padding}")
        left_pad = f"p{padding}" if padding > 0 else ""

        # actually encode the value
        coded = bitstruct.pack(f"{left_pad}{format_char}{bit_length}", raw_value)

        # create the raw mask of used bits for numeric objects
        used_mask_raw = used_mask

        if used_mask_raw is None:
            used_mask_raw = ((1 << bit_length) - 1).to_bytes((bit_length + 7) // 8, "big")

        if self.cursor_bit_position != 0:
            tmp = int.from_bytes(used_mask_raw, "big")
            tmp <<= self.cursor_bit_position
            used_mask_raw = tmp.to_bytes((self.cursor_bit_position + bit_length + 7) // 8, "big")

        # apply byte order to numeric objects
        if not is_highlow_byte_order and base_data_type in [
                DataType.A_INT32, DataType.A_UINT32, DataType.A_FLOAT32, DataType.A_FLOAT64
        ]:
            coded = coded[::-1]
            used_mask_raw = used_mask_raw[::-1]

        self.cursor_bit_position = 0
        self.emplace_bytes(coded, obj_used_mask=used_mask_raw)

    def emplace_bytes(self,
                      new_data: bytes,
                      obj_name: str | None = None,
                      obj_used_mask: bytes | None = None) -> None:
        if self.cursor_bit_position != 0:
            odxraise("EncodeState.emplace_bytes can only be called "
                     "for a bit position of 0!", RuntimeError)

        pos = self.cursor_byte_position

        # Make blob longer if necessary
        min_length = pos + len(new_data)
        if len(self.coded_message) < min_length:
            pad = b'\x00' * (min_length - len(self.coded_message))
            self.coded_message += pad
            self.used_mask += pad

        if obj_used_mask is None:
            # Happy path for when no obj_used_mask has been
            # specified. In this case we assume that all bits of the
            # new data to be emplaced are used.
            n = len(new_data)

            if self.used_mask[pos:pos + n] != b'\x00' * n:
                warnings.warn(
                    f"Overlapping objects detected in between bytes {pos} and "
                    f"{pos+n}",
                    OdxWarning,
                    stacklevel=1,
                )
            self.coded_message[pos:pos + n] = new_data
            self.used_mask[pos:pos + n] = b'\xff' * n
        else:
            # insert data the hard way, i.e. we have to look at each
            # individual byte to determine if it has already been used
            # somewhere else (it would be nice if bytearrays supported
            # bitwise operations!)
            for i in range(len(new_data)):
                if self.used_mask[pos + i] & obj_used_mask[i] != 0:
                    warnings.warn(
                        f"Overlapping objects detected at position {pos + i}",
                        OdxWarning,
                        stacklevel=1,
                    )
                self.coded_message[pos + i] &= ~obj_used_mask[i]
                self.coded_message[pos + i] |= new_data[i] & obj_used_mask[i]
                self.used_mask[pos + i] |= obj_used_mask[i]

        self.cursor_byte_position += len(new_data)

    @staticmethod
    def __encode_bcd_p(value: int) -> int:
        result = 0
        shift = 0
        while value > 0:
            result |= (value % 10) << shift
            shift += 4
            value //= 10

        return result

    @staticmethod
    def __encode_bcd_up(value: int) -> int:
        result = 0
        shift = 0
        while value > 0:
            result |= (value % 10) << shift
            shift += 8
            value //= 10

        return result



# SPDX-License-Identifier: MIT
from xml.etree import ElementTree

from ..exceptions import odxraise
from ..globals import xsi
from ..odxdoccontext import OdxDocContext
from .codedconstparameter import CodedConstParameter
from .dynamicparameter import DynamicParameter
from .lengthkeyparameter import LengthKeyParameter
from .matchingrequestparameter import MatchingRequestParameter
from .nrcconstparameter import NrcConstParameter
from .parameter import Parameter
from .physicalconstantparameter import PhysicalConstantParameter
from .reservedparameter import ReservedParameter
from .systemparameter import SystemParameter
from .tableentryparameter import TableEntryParameter
from .tablekeyparameter import TableKeyParameter
from .tablestructparameter import TableStructParameter
from .valueparameter import ValueParameter


def create_any_parameter_from_et(et_element: ElementTree.Element,
                                 context: OdxDocContext) \
                                 -> Parameter:
    parameter_type = et_element.get(f"{xsi}type")

    # Which attributes are set depends on the type of the parameter.
    if parameter_type == "VALUE":
        return ValueParameter.from_et(et_element, context)
    elif parameter_type == "CODED-CONST":
        return CodedConstParameter.from_et(et_element, context)
    elif parameter_type == "PHYS-CONST":
        return PhysicalConstantParameter.from_et(et_element, context)
    elif parameter_type == "SYSTEM":
        return SystemParameter.from_et(et_element, context)
    elif parameter_type == "LENGTH-KEY":
        return LengthKeyParameter.from_et(et_element, context)
    elif parameter_type == "NRC-CONST":
        return NrcConstParameter.from_et(et_element, context)
    elif parameter_type == "RESERVED":
        return ReservedParameter.from_et(et_element, context)
    elif parameter_type == "MATCHING-REQUEST-PARAM":
        return MatchingRequestParameter.from_et(et_element, context)
    elif parameter_type == "DYNAMIC":
        return DynamicParameter.from_et(et_element, context)
    elif parameter_type == "TABLE-STRUCT":
        return TableStructParameter.from_et(et_element, context)
    elif parameter_type == "TABLE-KEY":
        return TableKeyParameter.from_et(et_element, context)
    elif parameter_type == "TABLE-ENTRY":
        return TableEntryParameter.from_et(et_element, context)

    odxraise(f"I don't know about parameters of type {parameter_type}", NotImplementedError)
    return Parameter.from_et(et_element, context)



from xml.etree import ElementTree

from .comparam import Comparam
from .complexcomparam import ComplexComparam
from .odxdoccontext import OdxDocContext


def create_any_comparam_from_et(et_element: ElementTree.Element,
                                context: OdxDocContext) -> Comparam | ComplexComparam:
    if et_element.tag == "COMPARAM":
        return Comparam.from_et(et_element, context)
    elif et_element.tag == "COMPLEX-COMPARAM":
        return ComplexComparam.from_et(et_element, context)

    raise RuntimeError(f"Unhandled communication parameter type {et_element.tag}")



# SPDX-License-Identifier: MIT
from copy import deepcopy
from dataclasses import dataclass
from typing import Any, cast
from xml.etree import ElementTree

from ..comparamspec import ComparamSpec
from ..exceptions import odxassert
from ..odxdoccontext import OdxDocContext
from ..protstack import ProtStack
from .hierarchyelement import HierarchyElement
from .protocolraw import ProtocolRaw


@dataclass(kw_only=True)
class Protocol(HierarchyElement):
    """This is the class for primitives that are common for a given communication protocol

    Most importantly this diagnostic layer is responsible for defining
    the communication parameters that ought to be used.
    """

    @property
    def protocol_raw(self) -> ProtocolRaw:
        return cast(ProtocolRaw, self.diag_layer_raw)

    @property
    def comparam_spec(self) -> ComparamSpec:
        return self.protocol_raw.comparam_spec

    @property
    def prot_stack(self) -> ProtStack | None:
        return self.protocol_raw.prot_stack

    @staticmethod
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "Protocol":
        protocol_raw = ProtocolRaw.from_et(et_element, context)

        return Protocol(diag_layer_raw=protocol_raw)

    def __post_init__(self) -> None:
        super().__post_init__()

        odxassert(
            isinstance(self.diag_layer_raw, ProtocolRaw),
            "The raw diagnostic layer passed to Protocol "
            "must be a ProtocolRaw")

    def __deepcopy__(self, memo: dict[int, Any]) -> Any:
        """Create a deep copy of the protocol layer

        Note that the copied diagnostic layer is not fully
        initialized, so `_finalize_init()` should to be called on it
        before it can be used normally.
        """

        result = super().__deepcopy__(memo)

        # note that the self.protocol_raw object is *not* copied at
        # this place because the attribute points to the same object
        # as self.diag_layer_raw.
        result.protocol_raw = deepcopy(self.protocol_raw, memo)

        return result



# SPDX-License-Identifier: MIT
import argparse
import logging
import sys
from typing import cast

import InquirerPy.prompt as IP_prompt

from ..complexdop import ComplexDop
from ..database import Database
from ..dataobjectproperty import DataObjectProperty
from ..diaglayer import DiagLayer
from ..diagservice import DiagService
from ..dopbase import DopBase
from ..exceptions import odxraise, odxrequire
from ..hierarchyelement import HierarchyElement
from ..odxlink import resolve_snref
from ..odxtypes import AtomicOdxType, DataType, ParameterValueDict
from ..parameters.matchingrequestparameter import MatchingRequestParameter
from ..parameters.parameter import Parameter
from ..parameters.parameterwithdop import ParameterWithDOP
from ..parameters.valueparameter import ValueParameter
from ..request import Request
from ..response import Response
from . import _parser_utils
from ._parser_utils import SubparsersList
from ._print_utils import extract_parameter_tabulation_data

# name of the tool
_odxtools_tool_name_ = "browse"


def _convert_string_to_odx_type(string_value: str, odx_type: DataType) -> AtomicOdxType:
    """Similar to odx_type.from_string(string_value) but more relaxed to parse user input"""
    if odx_type == DataType.A_UINT32:
        return int(string_value, 0)
    elif odx_type == DataType.A_BYTEFIELD:
        return _convert_string_to_bytes(string_value)
    else:
        return odx_type.from_string(string_value)


def _convert_string_to_bytes(string_value: str) -> bytes:
    if all(len(x) <= 2 for x in string_value.split(" ")):
        return bytes(int(x, 16) for x in string_value.split(" ") if len(x) > 0)
    else:
        return int(string_value, 16).to_bytes((int(string_value, 16).bit_length() + 7) // 8, "big")


def _validate_string_value(input: str, parameter: Parameter) -> bool:
    if not parameter.is_required and input == "":
        return True
    elif isinstance(parameter, ParameterWithDOP):
        try:
            phys_type = odxrequire(parameter.physical_type)
            val = _convert_string_to_odx_type(input, phys_type.base_data_type)
        except:  # noqa: E722
            return False
        dop = parameter.dop
        if isinstance(dop, DataObjectProperty):
            return dop.is_valid_physical_value(val)
        else:
            raise NotImplementedError("Validation of complex DOPs")
    else:
        logging.info("This value is not validated precisely: Parameter {parameter}")
        return input != ""


def prompt_single_parameter_value(parameter: Parameter) -> AtomicOdxType | None:
    if not isinstance(parameter, ValueParameter):
        odxraise("Only the value of ValueParameters can be queried")
    if parameter.physical_type is None:
        odxraise("Only ValueParameters which define a physical data type can be queried")

    # TODO: add valid choices for the parameter
    #        "choices": parameter.get_valid_physical_values(),
    param_prompt = [{
        "type":
            "input",
        "name":
            parameter.short_name,
        "message":
            f"Value for parameter '{parameter.short_name}' (Type: {parameter.physical_type.base_data_type})"
            + (f"[optional]" if not parameter.is_required else ""),
        # TODO: improve validation
        "validate":
            lambda x: _validate_string_value(x, parameter),
        # TODO: do type conversion?
        "filter":
            lambda x: x
        # x if x == "" or p.physical_type.base_data_type is None
        # else _convert_string_to_odx_type(x, p.physical_type.base_data_type, param=p) # This does not work because the next parameter to be promted is used (for some reason?)
    }]

    if (dop := getattr(parameter, "dop", None)) and \
       (compu_method := getattr(dop, "compu_method", None)):
        scales = compu_method.internal_to_phys
        choices = [scale.compu_const for scale in scales if scale is not None]
        if (cdv := compu_method.compu_default_value) is not None:
            choices.append(cdv.compu_const)
        param_prompt[0]["choices"] = choices

    answer = IP_prompt(param_prompt)
    if answer.get(parameter.short_name) == "" and not parameter.is_required:
        return None
    elif parameter.physical_type.base_data_type is not None:
        return _convert_string_to_odx_type(
            cast(str, answer.get(parameter.short_name)), parameter.physical_type.base_data_type)
    else:
        logging.warning(
            f"Parameter {parameter.short_name} does not have a physical data type. Param details: {parameter}"
        )
        return cast(str, answer.get(parameter.short_name))


def encode_message_interactively(codec: Request | Response,
                                 ask_user_confirmation: bool = False) -> None:
    if sys.__stdin__ is None or sys.__stdout__ is None or not sys.__stdin__.isatty(
    ) or not sys.__stdout__.isatty():
        raise SystemError("This command can only be used in an interactive shell!")

    answered_request = b''
    if isinstance(codec, Response):
        answered_request_prompt = [{
            "type":
                "input",
            "name":
                "request",
            "message":
                f"What is the request you want to answer? (Enter the coded request as integer in hexadecimal format (e.g. 12 3B 5)",
            "filter":
                lambda input: _convert_string_to_bytes(input),
        }]
        answer = IP_prompt(answered_request_prompt)
        answered_request = cast(bytes, answer.get("request"))
        print(f"Input interpretation as list: {list(answered_request)}")

    has_settable_param = False
    for param in codec.parameters:
        if not param.is_settable:
            continue

        # TODO: Specifying complex parameters with nesting depth > 1
        # is not possible yet
        if (inner_params := getattr(getattr(param, "dop", None), "parameters", None)) is not None:
            for inner_param in inner_params:
                if inner_param.is_settable:
                    has_settable_param = True
        elif param.is_settable:
            has_settable_param = True

    param_values: ParameterValueDict = {}
    if has_settable_param:
        # Ask whether user wants to encode a message
        if ask_user_confirmation:
            encode_message_prompt = [{
                "type": "list",
                "name": "yes_no_prompt",
                "message": f"Do you want to encode a message? [y/n]",
                "choices": ["yes", "no"],
            }]
            answer = IP_prompt(encode_message_prompt)
            if answer.get("yes_no_prompt") == "no":
                return

        # Query user for the values of all settable parameters
        for param in codec.parameters:
            if (inner_params := getattr(dop := getattr(param, "dop", None), "parameters",
                                        None)) is not None:
                assert isinstance(dop, DopBase)
                inner_params = cast(list[Parameter], inner_params)
                # param refers to a complex DOP, i.e., the required
                # value is a key-value dict
                print(
                    f"The next {len(inner_params)} parameters belong to the structure '{dop.short_name}'"
                )
                structure_param_values: ParameterValueDict = {}
                for inner_param in inner_params:
                    if inner_param.is_settable:
                        val = prompt_single_parameter_value(inner_param)
                        if val is not None:
                            structure_param_values[inner_param.short_name] = val
                param_values[param.short_name] = structure_param_values
            elif param.is_settable:
                val = prompt_single_parameter_value(param)
                if val is not None:
                    param_values[param.short_name] = val

        if isinstance(codec, Response):
            payload = codec.encode(coded_request=answered_request, **param_values)
        else:
            payload = codec.encode(coded_request=b'', **param_values)
    else:
        # There are no settable parameters -> Just print message
        if isinstance(codec, Response):
            payload = codec.encode(coded_request=answered_request)
        else:
            payload = codec.encode()

    print(f"Message payload: 0x{bytes(payload).hex()}")


def encode_message_from_string_values(
    sub_service: Request | Response,
    parameter_values: ParameterValueDict | None = None,
) -> None:
    if parameter_values is None:
        parameter_values = {}
    parameter_values = parameter_values.copy()

    # Check if all needed parameters have been specified
    missing_parameter_names = []
    for param in sub_service.parameters:
        if (inner_params := getattr(dop := getattr(param, "dop", None), "parameters",
                                    None)) is not None:
            inner_param_values = parameter_values.get(param.short_name, {})
            if not isinstance(inner_param_values, dict):
                print(f"Value for composite parameter {param.short_name} must be "
                      f"a dictionary, got {type(inner_param_values).__name__}")
                continue
            for inner_param in inner_params:
                if inner_param.is_required and inner_param.short_name not in inner_param_values:
                    missing_parameter_names.append(f"{param.short_name}.{inner_param.short_name}")
        else:
            if param.is_required and parameter_values.get(param.short_name) is None:
                missing_parameter_names.append(param.short_name)

    if len(missing_parameter_names) > 0:
        print("The following parameters are required but missing:")
        print(" - " + "\n - ".join(sorted(missing_parameter_names)))
        return

    # Request values for parameters
    for parameter_sn, parameter_value in parameter_values.items():
        parameter = resolve_snref(parameter_sn, sub_service.parameters, Parameter)
        if parameter is None:
            print(f"I don't know the parameter {parameter_sn}")
            continue

        if isinstance(parameter_value, dict):
            # parameter_value refers to a structure (represented as dict of params)
            dop = getattr(parameter, "dop", None)
            inner_params = getattr(dop, "parameters", None)
            assert isinstance(dop, ComplexDop)
            assert isinstance(inner_params, list)
            inner_params = cast(list[Parameter], inner_params)

            typed_dict = parameter_value.copy()
            for inner_param_sn, inner_param_value in parameter_value.items():
                inner_param = resolve_snref(inner_param_sn, inner_params, Parameter)
                if inner_param is None:
                    print(f"Unknown sub-parameter {inner_param_sn}")
                    continue
                if not isinstance(inner_param_value, str):
                    print(f"The value specified for parameter {inner_param_sn} is not a string")
                    continue

                typed_dict[inner_param_sn] = _convert_string_to_odx_type(
                    inner_param_value, inner_param.physical_type.base_data_type)
            parameter_values[parameter.short_name] = typed_dict
        else:
            if not isinstance(parameter_value, str):
                print(f"Value for parameter {parameter_sn} is not a string")
                continue

            if not isinstance(parameter, MatchingRequestParameter):
                parameter_values[parameter_sn] = _convert_string_to_odx_type(
                    parameter_value,
                    parameter.physical_type.base_data_type,  # type: ignore[attr-defined]
                )
            else:
                parameter_values[parameter_sn] = _convert_string_to_odx_type(
                    parameter_value, DataType.A_BYTEFIELD)

    payload = sub_service.encode(coded_request=b'\xff' * 100, **parameter_values)

    print(f"Message payload: 0x{bytes(payload).hex()}")


def browse(odxdb: Database) -> None:
    if sys.__stdin__ is None or sys.__stdout__ is None or not sys.__stdin__.isatty(
    ) or not sys.__stdout__.isatty():
        raise SystemError("This command can only be used in an interactive shell!")
    dl_names = [dl.short_name for dl in odxdb.diag_layers]
    while True:
        # Select an ECU
        selection = [{
            "type": "list",
            "name": "variant",
            "message": "Select a Variant.",
            "choices": list(dl_names) + ["[exit]"],
        }]
        answer = IP_prompt(selection)
        if answer.get("variant") == "[exit]":
            return

        variant_name = answer.get("variant")
        assert isinstance(variant_name, str)
        variant = odxdb.diag_layers[variant_name]
        print(f"{type(answer.get('variant'))=}")
        assert isinstance(variant, DiagLayer)

        if isinstance(variant, HierarchyElement):
            if (rx_id := variant.get_receive_id()) is not None:
                recv_id = hex(rx_id)
            else:
                recv_id = "None"

            if (tx_id := variant.get_send_id()) is not None:
                send_id = hex(tx_id)
            else:
                send_id = "None"

            print(
                f"{variant.variant_type.value} '{variant.short_name}' (Receive ID: {recv_id}, Send ID: {send_id})"
            )

        while True:
            services: list[DiagService] = [
                s for s in variant.services if isinstance(s, DiagService)
            ]
            # Select a service of the ECU
            selection = [{
                "type":
                    "list",
                "name":
                    "service",
                "message":
                    f"The variant {variant.short_name} offers the following services. Select one!",
                "choices": [s.short_name for s in services] + ["[back]"],
            }]
            answer = IP_prompt(selection)
            if answer.get("service") == "[back]":
                break

            service_sn = answer.get("service")
            assert isinstance(service_sn, str)

            service = variant.services[service_sn]
            assert isinstance(service, DiagService)
            assert service.request is not None
            assert service.positive_responses is not None
            assert service.negative_responses is not None

            # Select a request/ response of the service
            selection = [{
                "type":
                    "list",
                "name":
                    "message_type",
                "message":
                    "This service offers the following messages.",
                "choices": [{
                    "name": f"Request: {service.request.short_name}",
                    "value": service.request,
                    "short": f"Request: {service.request.short_name}",
                }] + [{
                    "name": f"Positive response: {pr.short_name}",
                    "value": pr,
                    "short": f"Positive response: {pr.short_name}",
                } for pr in service.positive_responses] + [{
                    "name": f"Negative response: {nr.short_name}",
                    "value": nr,
                    "short": f"Negative response: {nr.short_name}",
                } for nr in service.negative_responses] + ["[back]"],  # type: ignore
            }]
            answer = IP_prompt(selection)
            if answer.get("message_type") == "[back]":
                continue

            codec = answer.get("message_type")
            if codec is not None:
                assert isinstance(codec, (Request, Response))
                table = extract_parameter_tabulation_data(codec.parameters)
                print(table)

                encode_message_interactively(codec, ask_user_confirmation=True)


def add_subparser(subparsers: SubparsersList) -> None:
    # Browse interactively to avoid spamming the console.
    parser = subparsers.add_parser(
        "browse",
        description="Interactively browse the content of automotive diagnostic files (*.pdx).",
        help="Interactively browse the content of automotive diagnostic files.",
        formatter_class=argparse.RawTextHelpFormatter,
    )

    _parser_utils.add_pdx_argument(parser)


def run(args: argparse.Namespace) -> None:
    odxdb = _parser_utils.load_file(args)
    browse(odxdb)
